//------------------------------------------------------------------------------
// SegwayBase.cs
//
// Author: Ben Axelrod
// Date: April 2007
// Version 1.0
//
//------------------------------------------------------------------------------
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Xml;
using SegwayNativeWrapper;
using submgr = Microsoft.Dss.Services.SubscriptionManager;
using W3C.Soap;


/*
 * VARIABLE                         RMP 100/50              RMP200/400
 * Pitch                            7.8 counts/ deg         7.8 counts/ deg
 * Pitch Rate                       7.8 counts/ deg/sec     7.8 counts/ deg/sec
 * Roll Angle                       7.8 counts/ deg         7.8 counts/ deg
 * Roll Rate                        7.8 counts/ deg/sec     7.8 counts/ deg/sec
 * Left Wheel Speed                 401 counts/ m/sec       332 counts/ m/sec
 * Right Wheel Speed                401 counts/ m/sec       332 counts/ m/sec
 * Yaw Rate                         7.8 counts/ deg/sec     7.8 counts/ deg/sec
 * Servo Frame counter              0.01 sec/ frame         0.01 sec/ frame
 * Integrated Left Wheel Position   40181 counts/ m         33215 counts/ m
 * Integrated Right Wheel Position  40181 counts/ m         33215 counts/ m
 * Integrated Fore/aft position     40181 counts/ m         33215 counts/ m
 * Integrated Turn position         117031 counts/ rev      112644 counts/ rev
 * Powerbase Battery Voltage        4 counts/Volt           4 counts/Volt
 * User Interface Battery Voltage   1.4 + counts * 0.0125   1.4 + counts * 0.0125
 * Left Motor Torque                1463 counts/Nm          1094 counts/Nm
 * Right Motor Torque               1463 counts/Nm          1094 counts/Nm
 */

//RMP 200
//tire diameter = 48 cm
//Wheel Track Width  = 53 cm


namespace Robotics.SegwayBase
{
    
    /// <summary>
    /// The SegwayBase Service
    /// </summary>
    [DisplayName("SegwayBase")]
    [Description("The SegwayBase Service")]
    [Contract(Contract.Identifier)]
    public class SegwayBaseService : DsspServiceBase
    {
        /// <summary>
        /// _state
        /// </summary>
        private SegwayBaseState _state = new SegwayBaseState();

        private Port<DateTime> _timerPort = new Port<DateTime>();
        private SegwayDataUpdateOperations _internalPort = new SegwayDataUpdateOperations();

        SegwayNativeWrapperClass _segway = new SegwayNativeWrapperClass();

        //send a velocity command every ControlMs milliseconds
        //should not be more than 100 Hz or less than 2.5 Hz
        private readonly int ControlMs = 50; 

        /// <summary>
        /// _main Port
        /// </summary>
        [ServicePort("/SegwayBase", AllowMultipleInstances=false)]
        private SegwayBaseOperations _mainPort = new SegwayBaseOperations();

        [Partner("SubMgr", Contract = submgr.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.CreateAlways)]
        submgr.SubscriptionManagerPort _subMgrPort = new submgr.SubscriptionManagerPort();

        /// <summary>
        /// Default Service Constructor
        /// </summary>
        public SegwayBaseService(DsspServiceCreationPort creationPort) : 
                base(creationPort)
        {
        }

        /// <summary>
        /// Service Start
        /// </summary>
        protected override void Start()
        {
            //initialize state
            _state.SetVelocityCommand = 0;
            _state.SetTurnCommand = 0;
            
            // Listen on the main port for requests and call the appropriate handler.
            Interleave mainInterleave = ActivateDsspOperationHandlers();

            // Publish the service to the local Node Directory
            DirectoryInsert();

            // display HTTP service Uri
            LogInfo(LogGroups.Console, "Service uri: ");

            // Set up an interleave for notifications from simrobot
            mainInterleave.CombineWith(new Interleave(
                new ExclusiveReceiverGroup(
                    Arbiter.Receive<ReplaceMessageData1>(true, _internalPort, ReplaceMessageData1Handler),
                    Arbiter.Receive<ReplaceMessageData2>(true, _internalPort, ReplaceMessageData2Handler),
                    Arbiter.Receive<ReplaceMessageData3>(true, _internalPort, ReplaceMessageData3Handler),
                    Arbiter.Receive<ReplaceMessageData4>(true, _internalPort, ReplaceMessageData4Handler),
                    Arbiter.Receive<ReplaceMessageData5>(true, _internalPort, ReplaceMessageData5Handler),
                    Arbiter.Receive<ReplaceMessageData6>(true, _internalPort, ReplaceMessageData6Handler),
                    Arbiter.Receive<ReplaceMessageData7>(true, _internalPort, ReplaceMessageData7Handler)
                ),
                new ConcurrentReceiverGroup())
            );


            if (_segway.Init() == 0)
            {
                LogInfo(LogGroups.Console, "Segway Connected");
                
                //setup event handlers for usb data
                _segway.OnDataUpdate1 += new DataUpdate1(_segway_OnDataUpdate1);
                _segway.OnDataUpdate2 += new DataUpdate2(_segway_OnDataUpdate2);
                _segway.OnDataUpdate3 += new DataUpdate3(_segway_OnDataUpdate3);
                _segway.OnDataUpdate4 += new DataUpdate4(_segway_OnDataUpdate4);
                _segway.OnDataUpdate5 += new DataUpdate5(_segway_OnDataUpdate5);
                _segway.OnDataUpdate6 += new DataUpdate6(_segway_OnDataUpdate6);
                _segway.OnDataUpdate7 += new DataUpdate7(_segway_OnDataUpdate7);

                //send configuration data
                //put into balance mode
                //start receiving data
                SpawnIterator(ConfigureSegway);
            }
            else
            {
                LogError(LogGroups.Console, "SegwayBase Initialization Failed");
            }
                      
        }

        //send configuration data to segway
        IEnumerator<ITask> ConfigureSegway()
        {
            _segway.SetGainSchedule(0);

            System.Threading.Thread.Sleep(50);
            //yield return Arbiter.Receive(false, TimeoutPort(50), delegate(DateTime t) { });

            _segway.SetMaxAccScale(0.5);

            System.Threading.Thread.Sleep(50);
            //yield return Arbiter.Receive(false, TimeoutPort(50), delegate(DateTime t) { });

            _segway.SetMaxVelScale(0.5);

            System.Threading.Thread.Sleep(50);
            //yield return Arbiter.Receive(false, TimeoutPort(50), delegate(DateTime t) { });

            _segway.SetMaxTurnScale(0.5);

            System.Threading.Thread.Sleep(50);
            //yield return Arbiter.Receive(false, TimeoutPort(50), delegate(DateTime t) { });

            _segway.SetCurrentLimitScale(1.0);

            System.Threading.Thread.Sleep(50);
            //yield return Arbiter.Receive(false, TimeoutPort(50), delegate(DateTime t) { });

            //_segway.SetOperationModeBalance();

            //System.Threading.Thread.Sleep(50);
            //yield return Arbiter.Receive(false, TimeoutPort(50), delegate(DateTime t) { });

            _segway.ResetAllIntegrators();
            //_mainPort.Post(new ResetIntegrator(15));

            System.Threading.Thread.Sleep(50);
            //yield return Arbiter.Receive(false, TimeoutPort(50), delegate(DateTime t) { });

            //_segway.ResetAllIntegrators();

            //System.Threading.Thread.Sleep(50);
            //yield return Arbiter.Receive(false, TimeoutPort(50), delegate(DateTime t) { });

            //start sending periodic commands
            _timerPort.Post(DateTime.Now);
            Activate(Arbiter.Receive(true, _timerPort, TimerHandler));

            System.Threading.Thread.Sleep(50);
            //yield return Arbiter.Receive(false, TimeoutPort(50), delegate(DateTime t) { });

            //start receiving usb data
            //Spawn(DoOnce);
            _segway.getUSBData();  //never returns
            yield break; 
        }


        #region USB Event Handlers and state updaters


        //event handlers
        void _segway_OnDataUpdate1(MessageData1 m) { _internalPort.Post(new ReplaceMessageData1(m)); }
        void _segway_OnDataUpdate2(MessageData2 m) { _internalPort.Post(new ReplaceMessageData2(m)); }
        void _segway_OnDataUpdate3(MessageData3 m) { _internalPort.Post(new ReplaceMessageData3(m)); }
        void _segway_OnDataUpdate4(MessageData4 m) { _internalPort.Post(new ReplaceMessageData4(m)); }
        void _segway_OnDataUpdate5(MessageData5 m) { _internalPort.Post(new ReplaceMessageData5(m)); }
        void _segway_OnDataUpdate6(MessageData6 m) { _internalPort.Post(new ReplaceMessageData6(m)); }
        void _segway_OnDataUpdate7(MessageData7 m) { _internalPort.Post(new ReplaceMessageData7(m)); }

        void ReplaceMessageData1Handler(ReplaceMessageData1 msg)
        {
            //update state
            _state.PitchAngle = msg.Body.PitchAngle;
            _state.PitchRate = msg.Body.PitchRate;
            _state.RollAngle = msg.Body.RollAngle;
            _state.RollRate = msg.Body.RollRate;

            _state.TimeStamp = DateTime.Now;

            // notify general subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest));

            // notify selective subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest, new string[] { "1" }));
        }

        void ReplaceMessageData2Handler(ReplaceMessageData2 msg)
        {
            _state.LeftWheelVelocity = msg.Body.LeftWheelVelocity;
            _state.RightWheelVelocity = msg.Body.RightWheelVelocity;
            _state.ServoFrames = msg.Body.ServoFrames;
            _state.YawRate = msg.Body.YawRate;

            _state.TimeStamp = DateTime.Now;

            // notify general subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest));

            // notify selective subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest, new string[] { "2" }));
        }


        void ReplaceMessageData3Handler(ReplaceMessageData3 msg)
        {
            _state.LeftWheelDisplacement = msg.Body.LeftWheelDisplacement;
            _state.RightWheelDisplacement = msg.Body.RightWheelDisplacement;

            _state.TimeStamp = DateTime.Now;

            // notify general subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest));

            // notify selective subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest, new string[] { "3" }));
        }

        void ReplaceMessageData4Handler(ReplaceMessageData4 msg)
        {
            _state.ForeAftDisplacement = msg.Body.ForeAftDisplacement;
            _state.YawDisplacement = msg.Body.YawDisplacement;

            _state.TimeStamp = DateTime.Now;

            // notify general subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest));

            // notify selective subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest, new string[] { "4" }));
        }

        void ReplaceMessageData5Handler(ReplaceMessageData5 msg)
        {
            _state.LeftMotorTorque = msg.Body.LeftMotorTorque;
            _state.RightMotorTorque = msg.Body.RightMotorTorque;

            _state.TimeStamp = DateTime.Now;

            // notify general subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest));

            // notify selective subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest, new string[] { "5" }));
        }

        void ReplaceMessageData6Handler(ReplaceMessageData6 msg)
        {
            _state.ControllerGainSchedule = msg.Body.ControllerGainSchedule;
            _state.OperationMode = msg.Body.OperationMode;
            _state.PBBatteryVoltage = msg.Body.PBBatteryVoltage;
            _state.UIBatteryVoltage = msg.Body.UIBatteryVoltage;

            _state.TimeStamp = DateTime.Now;

            // notify general subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest));

            // notify selective subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest, new string[] { "6" }));
        }

        void ReplaceMessageData7Handler(ReplaceMessageData7 msg)
        {
            _state.ReadTurnCommand = msg.Body.TurnCommand;
            _state.ReadVelocityCommand = msg.Body.VelocityCommand;

            _state.TimeStamp = DateTime.Now;

            // notify general subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest));

            // notify selective subscribers
            _subMgrPort.Post(new submgr.Submit(_state, DsspActions.ReplaceRequest, new string[] { "7" }));
        }
        
        #endregion




        ///// <summary>
        ///// spawns a thread to handle all USB comms.  
        ///// never returns.
        ///// </summary>
        //void DoOnce()
        //{
        //    _segway.getUSBData();
        //}
        

        void TimerHandler(DateTime signal)
        {
            _segway.Drive(_state.SetVelocityCommand, _state.SetTurnCommand);

            Activate(
                Arbiter.Receive(false, TimeoutPort(ControlMs),
                    delegate(DateTime time)
                    {
                        _timerPort.Post(time);
                    }
                )
            );
        }


        /// <summary>
        /// Get Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        /// <summary>
        /// Replace Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReplaceHandler(Replace replace)
        {
            _state = replace.Body;
            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);
            yield break;
        }

        /// <summary>
        /// Drive Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> DriveHandler(Drive drive)
        {
            
            _state.SetVelocityCommand = drive.Body.Velocity;
            _state.SetTurnCommand = drive.Body.Turn;

            //Console.WriteLine(_state.SetVelocityCommand + "   " + _state.SetTurnCommand);

            _segway.Drive(_state.SetVelocityCommand, _state.SetTurnCommand);
            drive.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

        /// <summary>
        /// reset integrator Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ResetIntegratorHandler(ResetIntegrator reset)
        {
            Console.WriteLine("encoders before: " + _state.LeftWheelDisplacement + "  " + _state.RightWheelDisplacement);
            _segway.ResetIntegrator(reset.Body.Which);
            Console.WriteLine("encoders after: " + _state.LeftWheelDisplacement + "  " + _state.RightWheelDisplacement);
            reset.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

        /// <summary>
        /// General Subscription
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> SubscribeHandler(Subscribe subscribe)
        {
            base.SubscribeHelper(_subMgrPort, subscribe.Body, subscribe.ResponsePort);
            yield break;
        }

        /// <summary>
        /// Custom Subscription
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> SelectiveSubscribeHandler(SelectiveSubscribe subRequest)
        {
            submgr.InsertSubscription selectiveSubscription = 
                new submgr.InsertSubscription(
                new submgr.InsertSubscriptionMessage(subRequest.Body.Subscriber, subRequest.Body.Expiration, 0));

            selectiveSubscription.Body.NotificationCount = subRequest.Body.NotificationCount;

            List<submgr.QueryType> subscribeFilter = new List<submgr.QueryType>();

            //items in this loop are OR'ed together in the subscription
            foreach (int i in subRequest.Body.DataMessageNumbers)
            {
                LogInfo("Segway base adding subscription for DataMessage" + i);

                //you can achieve an AND behavior by adding a list of strings in the new QueryType
                subscribeFilter.Add(new submgr.QueryType(i.ToString()));
            }

            selectiveSubscription.Body.QueryList = subscribeFilter.ToArray();
            _subMgrPort.Post(selectiveSubscription);

            yield return Arbiter.Choice(selectiveSubscription.ResponsePort,
                delegate(SubscribeResponseType response)
                {
                    subRequest.ResponsePort.Post(response);
                },
                delegate(Fault fault)
                {
                    subRequest.ResponsePort.Post(fault);
                }
            );
            yield break;
        }

    }
}
