//------------------------------------------------------------------------------
// SegwayBaseTypes.cs
//
// Author: Ben Axelrod
// Date: April 2007
// Version 1.0
//
//------------------------------------------------------------------------------
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using System;
using System.Collections.Generic;
using W3C.Soap;
using SegwayNativeWrapper;
using System.ComponentModel;

namespace Robotics.SegwayBase
{
    
    /// <summary>
    /// SegwayBase Contract
    /// </summary>
    public sealed class Contract
    {
        /// The Unique Contract Identifier for the SegwayBase service
        public const String Identifier = "http://schemas.tempuri.org/2007/02/segwaybase.html";
    }
    /// <summary>
    /// SegwayBase State
    /// </summary>
    [DataContract()]
    public class SegwayBaseState
    {
        [DataMember]
        public DateTime TimeStamp;

        [DataMember]
        public int PitchAngle;
        [DataMember]
        public int PitchRate;
        [DataMember]
        public int RollAngle;
        [DataMember]
        public int RollRate;

        [DataMember]
        public int LeftWheelVelocity;
        [DataMember]
        public int RightWheelVelocity;
        [DataMember]
        public int YawRate;
        [DataMember]
        public int ServoFrames;

        [DataMember]
        public int LeftWheelDisplacement;
        [DataMember]
        public int RightWheelDisplacement;

        [DataMember]
        public int ForeAftDisplacement;
        [DataMember]
        public int YawDisplacement;

        [DataMember]
        public int LeftMotorTorque;
        [DataMember]
        public int RightMotorTorque;

        [DataMember]
        public int OperationMode;
        [DataMember]
        public int ControllerGainSchedule;
        [DataMember]
        public int UIBatteryVoltage;
        [DataMember]
        public int PBBatteryVoltage;

        /// <summary>
        /// velocity setting read from segway
        /// </summary>
        [DataMember]
        public int ReadVelocityCommand;

        /// <summary>
        /// turn setting read from segway
        /// </summary>
        [DataMember]
        public int ReadTurnCommand;

        /// <summary>
        /// velocity setting to send to segway
        /// </summary>
        [DataMember]
        public int SetVelocityCommand;

        /// <summary>
        /// turn setting to send to segway
        /// </summary>
        [DataMember]
        public int SetTurnCommand;
    }

    /// <summary>
    /// SegwayBase Operations
    /// </summary>
    public class SegwayBaseOperations : PortSet<
        DsspDefaultLookup, 
        DsspDefaultDrop, 
        Get,
        Replace,
        Drive,
        SelectiveSubscribe,//IMPORTANT: Because SelectiveSubscribe inherits from Subscribe, it must go on top.
        Subscribe,
        ResetIntegrator>  
    {
    }

    /// <summary>
    /// Get
    /// </summary>
    public class Get : Get<GetRequestType, PortSet<SegwayBaseState, Fault>>
    {
    }

    /// <summary>
    /// replace state
    /// </summary>
    public class Replace : Replace<SegwayBaseState, PortSet<DefaultReplaceResponseType, Fault>>
    {
    }

    /// <summary>
    /// Send a drive message to Segway
    /// </summary>
    [Description("Drives the segway")]
    public class Drive : Update<DriveMessage, PortSet<DefaultUpdateResponseType, Fault>>
    {
    }

    [DataContract]
    public class DriveMessage
    {
        [DataMember]
        public int Velocity;

        [DataMember]
        public int Turn;
    }

    /// <summary>
    /// The standard subscription
    /// </summary>
    public class Subscribe : Subscribe<SubscribeRequestType, PortSet<SubscribeResponseType, Fault>> { }

    /// <summary>
    /// The custom subscription
    /// </summary>
    public class SelectiveSubscribe : Subscribe<MySubscribeRequestType, PortSet<SubscribeResponseType, Fault>> { }

    /// <summary>
    /// body of SelectiveSubscribe.  can subscribe to one of the 7 data messages.
    /// can be eithe
    /// </summary>
    [DataContract]
    public class MySubscribeRequestType : SubscribeRequestType
    {
        /// <summary>
        /// The list of data messages to subscribe to.  1 through 7
        /// </summary>
        [DataMember]
        public List<int> DataMessageNumbers;
    }

    /// <summary>
    /// Resets a particular integrator on the segway. 
    /// 1 = RIGHT WHEEL DISPLACEMENT
    /// 2 = LEFT WHEEL DISPLACEMENT
    /// 4 = YAW DISPLACEMENT
    /// 8 = FORE/AFT DISPLACEMENT
    /// </summary>
    [Description("Resets a particular integrator on the segway")]
    public class ResetIntegrator : Update<ResetIntegratorBody, PortSet<DefaultUpdateResponseType, Fault>>
    {
        public ResetIntegrator() { }
        public ResetIntegrator(int which)
        {
            ResetIntegratorBody b = new ResetIntegratorBody();
            b.Which = which;
            this.Body = b;
        }
    }

    [DataContract]
    public class ResetIntegratorBody
    {
        [DataMember]
        public int Which;
    }


    #region SegwayDataUpdateOperations

    public class SegwayDataUpdateOperations : PortSet<
        ReplaceMessageData1,
        ReplaceMessageData2,
        ReplaceMessageData3,
        ReplaceMessageData4,
        ReplaceMessageData5,
        ReplaceMessageData6,
        ReplaceMessageData7
        >
    {
    }

    /// <summary>
    /// update state with MessageData1 from Segway
    /// </summary>
    public class ReplaceMessageData1 : Replace<MyMessageData1, PortSet<DefaultReplaceResponseType, Fault>>
    {
        public ReplaceMessageData1() { }
        public ReplaceMessageData1(MessageData1 msg)
        {
            this.Body.PitchAngle = msg.PitchAngle;
            this.Body.PitchRate = msg.PitchRate;
            this.Body.RollAngle = msg.RollAngle;
            this.Body.RollRate = msg.RollRate;
        }
    }

    [DataContract()]
    public class MyMessageData1
    {
        [DataMember]
        public int PitchAngle;
        [DataMember]
        public int PitchRate;
        [DataMember]
        public int RollAngle;
        [DataMember]
        public int RollRate;
    }

    /// <summary>
    /// update state with MessageData2 from Segway
    /// </summary>
    public class ReplaceMessageData2 : Replace<MyMessageData2, PortSet<DefaultReplaceResponseType, Fault>>
    {
        public ReplaceMessageData2() { }
        public ReplaceMessageData2(MessageData2 msg)
        {
            this.Body.LeftWheelVelocity = msg.LeftWheelVelocity;
            this.Body.RightWheelVelocity = msg.RightWheelVelocity;
            this.Body.ServoFrames = msg.ServoFrames;
            this.Body.YawRate = msg.YawRate;
        }
    }

    [DataContract()]
    public class MyMessageData2
    {
        [DataMember]
        public int LeftWheelVelocity;
        [DataMember]
        public int RightWheelVelocity;
        [DataMember]
        public int YawRate;
        [DataMember]
        public int ServoFrames;
    }

    /// <summary>
    /// update state with MessageData3 from Segway
    /// </summary>
    public class ReplaceMessageData3 : Replace<MyMessageData3, PortSet<DefaultReplaceResponseType, Fault>>
    {
        public ReplaceMessageData3() { }
        public ReplaceMessageData3(MessageData3 msg)
        {
            this.Body.LeftWheelDisplacement = msg.LeftWheelDisplacement;
            this.Body.RightWheelDisplacement = msg.RightWheelDisplacement;
        }
    }

    [DataContract()]
    public class MyMessageData3
    {
        [DataMember]
        public int LeftWheelDisplacement;
        [DataMember]
        public int RightWheelDisplacement;
    }


    /// <summary>
    /// update state with MessageData4 from Segway
    /// </summary>
    public class ReplaceMessageData4 : Replace<MyMessageData4, PortSet<DefaultReplaceResponseType, Fault>>
    {
        public ReplaceMessageData4() { }
        public ReplaceMessageData4(MessageData4 msg)
        {
            this.Body.ForeAftDisplacement = msg.ForeAftDisplacement;
            this.Body.YawDisplacement = msg.YawDisplacement;
        }
    }

    [DataContract()]
    public class MyMessageData4
    {
        [DataMember]
        public int ForeAftDisplacement;
        [DataMember]
        public int YawDisplacement;
    }

    /// <summary>
    /// update state with MessageData5 from Segway
    /// </summary>
    public class ReplaceMessageData5 : Replace<MyMessageData5, PortSet<DefaultReplaceResponseType, Fault>>
    {
        public ReplaceMessageData5() { }
        public ReplaceMessageData5(MessageData5 msg)
        {
            this.Body.LeftMotorTorque = msg.LeftMotorTorque;
            this.Body.RightMotorTorque = msg.RightMotorTorque;
        }
    }

    [DataContract()]
    public class MyMessageData5
    {
        [DataMember]
        public int LeftMotorTorque;
        [DataMember]
        public int RightMotorTorque;
    }

    /// <summary>
    /// update state with MessageData6 from Segway
    /// </summary>
    public class ReplaceMessageData6 : Replace<MyMessageData6, PortSet<DefaultReplaceResponseType, Fault>>
    {
        public ReplaceMessageData6() { }
        public ReplaceMessageData6(MessageData6 msg)
        {
            this.Body.ControllerGainSchedule = msg.ControllerGainSchedule;
            this.Body.OperationMode = msg.OperationMode;
            this.Body.PBBatteryVoltage = msg.PBBatteryVoltage;
            this.Body.UIBatteryVoltage = msg.UIBatteryVoltage;
        }
    }

    [DataContract()]
    public class MyMessageData6
    {
        [DataMember]
        public int OperationMode;
        [DataMember]
        public int ControllerGainSchedule;
        [DataMember]
        public int UIBatteryVoltage;
        [DataMember]
        public int PBBatteryVoltage;
    }

    /// <summary>
    /// update state with MessageData7 from Segway
    /// </summary>
    public class ReplaceMessageData7 : Replace<MyMessageData7, PortSet<DefaultReplaceResponseType, Fault>>
    {
        public ReplaceMessageData7() { }
        public ReplaceMessageData7(MessageData7 msg)
        {
            this.Body.TurnCommand = msg.TurnCommand;
            this.Body.VelocityCommand = msg.VelocityCommand;
        }
    }

    [DataContract()]
    public class MyMessageData7
    {
        [DataMember]
        public int VelocityCommand;
        [DataMember]
        public int TurnCommand;
    }

    #endregion

}
