// SegwayNativeWrapper.h
//
// Author: Ben Axelrod
// Date: April 2007
// Version 1.0
//
#pragma once

#include "usb_int.h"
#include "vcclr.h"

using namespace System;
using namespace System::Diagnostics;
using namespace System::Reflection;
using namespace System::Runtime::InteropServices;
using namespace System::Text;


//RMP message identifiers
//
#define	RID_COMMAND		0x0413	//RMP command
#define	RID_SHUTDOWN	0x0412	//RMP shutdown command
#define	RID_DATA0		0x0400	//Data from RMP
#define	RID_DATA1		0x0401
#define	RID_DATA2		0x0402
#define	RID_DATA3		0x0403
#define	RID_DATA4		0x0404
#define	RID_DATA5		0x0405
#define	RID_DATA6		0x0406
#define	RID_DATA7		0x0407
#define RID_HEARTBEAT	0x0688
#define RID_STATUS		0x0680

//RMP_CMD  are the code for the commands
//
#define	RMP_CMD_NONE			0
#define	RMP_CMD_RESET_TRAVEL	4
#define	RMP_CMD_ENTER_BALANCE	5


#define	RMP_CMD_SET_MAXIMUM_VELOCITY		10
#define	RMP_CMD_SET_MAXIMUM_ACCELERATION	11
#define	RMP_CMD_SET_MAXIMUM_TURN_RATE		12
#define RMP_CMD_SET_GAINMODE				13
#define RMP_CMD_SET_CURRENT_LIMIT			14
#define RMP_CMD_SET_BALANCE_LOCKOUT         15
#define RMP_CMD_SET_OPERATIONAL_MODE        16

#define	RMP_CMD_INTEGRATOR_RESET	50
#define	RMP_PARAM_RIGHT_WHEEL_POS	1
#define	RMP_PARAM_LEFT_WHEEL_POS	2
#define	RMP_PARAM_YAW_POS			4
#define	RMP_PARAM_FORE_AFT_POS		8
#define	RMP_PARAM_MR_ALL (\
							RMP_PARAM_RIGHT_WHEEL_POS | RMP_PARAM_LEFT_WHEEL_POS \
							| RMP_PARAM_YAW_POS |RMP_PARAM_FORE_AFT_POS )


#define RMP_OP_MODE_TRACTOR  1
#define RMP_OP_MODE_BALANCE  2
#define RMP_OP_MODE_OFF      3

#define RXMAX_BUFF_SIZE   8192 //4096


namespace SegwayNativeWrapper
{

	public value struct MessageData1
    {
		int PitchAngle;
		int PitchRate;
		int RollAngle;
		int RollRate;
	};

	public value struct MessageData2
    {
		int LeftWheelVelocity;
		int RightWheelVelocity;
		int YawRate;
		int ServoFrames;
	};

	public value struct MessageData3
    {
		public:
		int LeftWheelDisplacement;
		int RightWheelDisplacement;
	};

	public value struct MessageData4
    {
		public:
		int ForeAftDisplacement;
		int YawDisplacement;
	};

	public value struct MessageData5
    {
		public:
		int LeftMotorTorque;
		int RightMotorTorque;
	};

	public value struct MessageData6
    {
		public:
		int OperationMode;
		int ControllerGainSchedule;
		int UIBatteryVoltage;
		int PBBatteryVoltage;
	};

	public value struct MessageData7
    {
		public:
		int VelocityCommand;
		int TurnCommand;
    };

	public delegate void DataUpdate1(MessageData1 m);
	public delegate void DataUpdate2(MessageData2 m);
	public delegate void DataUpdate3(MessageData3 m);
	public delegate void DataUpdate4(MessageData4 m);
	public delegate void DataUpdate5(MessageData5 m);
	public delegate void DataUpdate6(MessageData6 m);
	public delegate void DataUpdate7(MessageData7 m);

	// ------------------------------------------------------------------------
	// ISegwayEvents Interface
	// ------------------------------------------------------------------------

	public interface class ISegwayEvents
	{
	public:
		void interopOnDataUpdate1(MessageData1);
		void interopOnDataUpdate2(MessageData2);
		void interopOnDataUpdate3(MessageData3);
		void interopOnDataUpdate4(MessageData4);
		void interopOnDataUpdate5(MessageData5);
		void interopOnDataUpdate6(MessageData6);
		void interopOnDataUpdate7(MessageData7);
	};

	public ref class SegwayNativeWrapperClass : public ISegwayEvents
	{

	private:

		int Velocity;
		int Turn;

		byte* rx;

	public:
		
		event DataUpdate1^ OnDataUpdate1;  // outbound to clients
		event DataUpdate2^ OnDataUpdate2;
		event DataUpdate3^ OnDataUpdate3;
		event DataUpdate4^ OnDataUpdate4;
		event DataUpdate5^ OnDataUpdate5;
		event DataUpdate6^ OnDataUpdate6;
		event DataUpdate7^ OnDataUpdate7;

		virtual void interopOnDataUpdate1(MessageData1 msg) { OnDataUpdate1(msg); };
		virtual void interopOnDataUpdate2(MessageData2 msg) { OnDataUpdate2(msg); };
		virtual void interopOnDataUpdate3(MessageData3 msg) { OnDataUpdate3(msg); };
		virtual void interopOnDataUpdate4(MessageData4 msg) { OnDataUpdate4(msg); };
		virtual void interopOnDataUpdate5(MessageData5 msg) { OnDataUpdate5(msg); };
		virtual void interopOnDataUpdate6(MessageData6 msg) { OnDataUpdate6(msg); };
		virtual void interopOnDataUpdate7(MessageData7 msg) { OnDataUpdate7(msg); };

		SegwayNativeWrapperClass()
		{
			rx = new byte[RXMAX_BUFF_SIZE];

			Velocity = 0;
			Turn = 0;
		};

		~SegwayNativeWrapperClass()
		{
			delete rx;
			usb_closedll();
		};
		
		//must call first
		int Init()
		{
			return usb_init();
		};

		//use with caution
		void Shutdown()
		{
			unsigned short junk[4];
			usb_can_send(1, RID_SHUTDOWN, junk);
		};


		//velocity
		//-1176 to 1176
		//(147 counts / mph  so:  -8 mph to 8 mph)
		//
		//turn
		//-1024 to 1024
		//(147 counts / mph  so:  -8 mph to 8 mph)
		void Drive(int vel, int turn)
		{
			Velocity = vel;
			Turn = turn;

			unsigned short data[4];
			data[0] = Velocity & 0xFFFF;
			data[1] = Turn & 0xFFFF;
			data[2] = RMP_CMD_NONE;
			data[3] = 0;
			usb_can_send(1, RID_COMMAND, data);
		};
		

		// 0->1 maps to 0 -> 16 
		void SetMaxVelScale(double velScale)
		{
			int scale = (int)System::Math::Round(16 * velScale);
			unsigned short data[4];
			data[0] = Velocity & 0xFFFF;
			data[1] = Turn & 0xFFFF;
			data[2] = RMP_CMD_SET_MAXIMUM_VELOCITY;
			data[3] = scale & 0x0F;
			usb_can_send(1, RID_COMMAND, data);
		};


		// 0->1 maps to 0 -> 16 
		void SetMaxAccScale(double accScale)
		{
			int scale = (int)System::Math::Round(16 * accScale);
			unsigned short data[4];
			data[0] = Velocity & 0xFFFF;
			data[1] = Turn & 0xFFFF;
			data[2] = RMP_CMD_SET_MAXIMUM_ACCELERATION;
			data[3] = scale & 0x0F;
			usb_can_send(1, RID_COMMAND, data);
		};

		// 0->1 maps to 0 -> 16 
		void SetMaxTurnScale(double turnRate)
		{
			int rate = (int)System::Math::Round(16 * turnRate);
			unsigned short data[4];
			data[0] = Velocity & 0xFFFF;
			data[1] = Turn & 0xFFFF;
			data[2] = RMP_CMD_SET_MAXIMUM_TURN_RATE;
			data[3] = rate & 0x0F;
			usb_can_send(1, RID_COMMAND, data);
		};


		/// <summary>
        /// 0, 1, 2
		/// NOTE: cannot be in balance mode to set
        /// </summary>
		void SetGainSchedule(int gain)
		{
			unsigned short data[4];
			data[0] = Velocity & 0xFFFF;
			data[1] = Turn & 0xFFFF;
			data[2] = RMP_CMD_SET_GAINMODE;
			data[3] = gain & 3;
			usb_can_send(1, RID_COMMAND, data);
		};
		
		///0->1 maps to 0 -> 256
		void SetCurrentLimitScale(double currLimit)
		{
			int lim = (int)System::Math::Round(256 * currLimit);
			unsigned short data[4];
			data[0] = Velocity & 0xFFFF;
			data[1] = Turn & 0xFFFF;
			data[2] = RMP_CMD_SET_CURRENT_LIMIT;
			data[3] = lim & 0xFF;
			usb_can_send(1, RID_COMMAND, data);
		};

		//true, false
		//NOTE: cannot use in balance mode
		void SetBalanceModeLockout(bool lockout)
		{
			unsigned short data[4];
			data[0] = Velocity & 0xFFFF;
			data[1] = Turn & 0xFFFF;
			data[2] = RMP_CMD_SET_BALANCE_LOCKOUT;
			data[3] = lockout ? 1 : 0;
			usb_can_send(1, RID_COMMAND, data);
		};
	
	private:
		//1, 2, 3
		void SetOperationMode(int mode)
		{
			unsigned short data[4];
			data[0] = Velocity & 0xFFFF;
			data[1] = Turn & 0xFFFF;
			data[2] = RMP_CMD_SET_OPERATIONAL_MODE;
			data[3] = mode & 0xff;
			usb_can_send(1, RID_COMMAND, data);
		};

	public:
		void SetOperationModeTractor() { SetOperationMode(RMP_OP_MODE_TRACTOR); };
		void SetOperationModeBalance() { SetOperationMode(RMP_OP_MODE_BALANCE); };
		void SetOperationModeOff()     { SetOperationMode(RMP_OP_MODE_OFF); };

		//1, 2, 4, 8
		void ResetIntegrator(int which)
		{
			unsigned short data[4];
			data[0] = Velocity & 0xFFFF;
			data[1] = Turn & 0xFFFF;
			data[2] = RMP_CMD_INTEGRATOR_RESET;
			data[3] = which & 0xF;
			usb_can_send(1, RID_COMMAND, data);
		};

		void ResetAllIntegrators()	     { ResetIntegrator(RMP_PARAM_MR_ALL); };
		void ResetWheelIntegrators()     { ResetIntegrator(RMP_PARAM_RIGHT_WHEEL_POS | RMP_PARAM_LEFT_WHEEL_POS); };
		void ResetLeftWheelIntegrator()  { ResetIntegrator(RMP_PARAM_LEFT_WHEEL_POS); };
		void ResetRightWheelIntegrator() { ResetIntegrator(RMP_PARAM_RIGHT_WHEEL_POS); };
		void ResetYawIntegrator()	     { ResetIntegrator(RMP_PARAM_YAW_POS); };
		void ResetForeAftIntegrator()    { ResetIntegrator(RMP_PARAM_FORE_AFT_POS); };
		
		void getUSBData()
		{

			DWORD ret_bytes;
			DWORD bytes_in_buf = 0;
			FT_STATUS status;
			static	int counter = 0;
			int	start_index = 0;	// offset into receive buffer
			int	bytes_remaining;
			unsigned int idx;

			status = usb_GetQueueStatus(&bytes_in_buf);
			if (bytes_in_buf > RXMAX_BUFF_SIZE)
			{
				//System::Console::WriteLine("!!!CLEARING SEGWAY BUFFER!!!");
				while(bytes_in_buf > 0)
				{
					if (bytes_in_buf > RXMAX_BUFF_SIZE)
						bytes_in_buf = RXMAX_BUFF_SIZE;
					status = usb_Read(&rx[0], bytes_in_buf, &ret_bytes);
					status = usb_GetQueueStatus(&bytes_in_buf);
				}
			}

			//loop forever...
			while (1)
			{
				status = usb_GetQueueStatus(&bytes_in_buf);
				if (status != FT_OK)  
				{
					bytes_in_buf = 0;
				}

				//System::Console::WriteLine((unsigned int)bytes_in_buf);
				
				/*if(bytes_in_buf == 0)
				{
					System::Threading::Thread::Sleep(1);
				}*/

				if ((bytes_in_buf + start_index) > 0)
				{

					if (bytes_in_buf > (RXMAX_BUFF_SIZE - start_index))
					{
						bytes_in_buf = RXMAX_BUFF_SIZE - start_index;
						System::Console::WriteLine("Segway buffer overflow");
					}

					status = usb_Read(&rx[start_index], bytes_in_buf, &ret_bytes);


					if (((status == FT_OK)  && (ret_bytes > 0)) ||
						((status == FT_IO_ERROR) && (ret_bytes > 0)) /* Read timeout */ )
					{
						//fprintf(xscript, "\nnewblock  len=%d\n", ret_bytes);
						idx = 0;

						bytes_remaining = ret_bytes + start_index;
						while (bytes_remaining >= 18)  {
							//find the first 'message'
							//
							while ((rx[idx] != 0x0f0) && (idx < ret_bytes))  
							{
								idx++;
								bytes_remaining--;
							}

							//fprintf(xscript, "-%02x-  %d", rx[idx], bytes_remaining);

							//each message should be 18 bytes in length
							//parse each one
							//
							if (bytes_remaining >= 18)  
							{
								parse_usb_message(&rx[idx]);
								bytes_remaining -= 18;
								idx += 18;
							}
						}

						if (bytes_remaining > 0)  
						{
							//fprintf(xscript, "hold %d\n", bytes_remaining);
							memcpy(&rx[0], &rx[idx], bytes_remaining);
							start_index = bytes_remaining;
						}
						else 
						{
							start_index = 0;
						}
					}
				}
				else
					System::Threading::Thread::Sleep(1);
			}
		};

		private:

		void parse_usb_message(unsigned char *bytes)
		{
			if (bytes[1] == 0)  
			{
				//this message is a HEARTBEAT
				//don't need to parse because we get battery info 
				//from RID_DATA6 message
			}
			else if (bytes[1] == 0x055 && bytes[2] == 0xaa)  
			{
				//bytes[2] ... 0xaa == side A,  0xbb == side B

				int id = ((bytes[4] << 3) | ((bytes[5] >> 5) & 7)) & 0x0fff;

				//int header = (bytes[6] << 8) | bytes[7];

				switch(id)
				{
					case RID_HEARTBEAT:
						//System::Console::Write(" RID_HEARTBEAT");
						break;
					case RID_SHUTDOWN:
						//System::Console::Write(" RID_SHUTDOWN");
						break;
					case RID_COMMAND:
						//System::Console::Write(" RID_COMMAND");
						break;
					case RID_STATUS:
						//System::Console::Write(" RID_STATUS");
						break;
					case RID_DATA0:
						//System::Console::Write(" RID_DATA0");
						//nothing to do
						break;
					case RID_DATA1:
						//System::Console::Write(" RID_DATA1  ");
						MessageData1 data1;
						data1.PitchAngle = GetShort(bytes, 9);
						data1.PitchRate  = GetShort(bytes, 11);
						data1.RollAngle  = GetShort(bytes, 13);
						data1.RollRate   = GetShort(bytes, 15);
						OnDataUpdate1(data1);
						break;
					case RID_DATA2:
						//System::Console::Write(" RID_DATA2");
						MessageData2 data2;
						data2.LeftWheelVelocity  = GetShort(bytes, 9);
						data2.RightWheelVelocity = GetShort(bytes, 11);
						data2.YawRate            = GetShort(bytes, 13);
						data2.ServoFrames        = GetShort(bytes, 15);
						OnDataUpdate2(data2);
						break;
					case RID_DATA3:
						//System::Console::Write(" RID_DATA3");
						MessageData3 data3;
						data3.LeftWheelDisplacement  = GetInt(bytes, 9);
						data3.RightWheelDisplacement = GetInt(bytes, 13);
						OnDataUpdate3(data3);
						break;
					case RID_DATA4:
						//System::Console::Write(" RID_DATA4");
						MessageData4 data4;
						data4.ForeAftDisplacement = GetInt(bytes, 9);
						data4.YawDisplacement     = GetInt(bytes, 13);
						OnDataUpdate4(data4);
						break;
					case RID_DATA5:
						//System::Console::Write(" RID_DATA5");
						MessageData5 data5;
						data5.LeftMotorTorque  = GetShort(bytes, 9);
						data5.RightMotorTorque = GetShort(bytes, 11);
						OnDataUpdate5(data5);
						break;
					case RID_DATA6:
						//System::Console::Write(" RID_DATA6");
						MessageData6 data6;
						data6.OperationMode          = GetShort(bytes, 9);
						data6.ControllerGainSchedule = GetShort(bytes, 11);
						data6.UIBatteryVoltage       = GetShort(bytes, 13);
						data6.PBBatteryVoltage       = GetShort(bytes, 15);
						OnDataUpdate6(data6);
						break;
					case RID_DATA7:
						//System::Console::Write(" RID_DATA7");
						MessageData7 data7;
						data7.VelocityCommand = GetShort(bytes, 9);
						data7.TurnCommand     = GetShort(bytes, 11);
						OnDataUpdate7(data7);
						break;
					default:
						System::Console::WriteLine("Unknown Segway USB Message");
						break;
				}

			}
		};


        int GetShort(unsigned char* data, int startPosition)
        {
			int temp = (data[startPosition] << 8) | data[startPosition+1];
			
			if (temp >= 0x8000)
			{
				temp = temp ^ 0xFFFF;
				temp = -(temp + 1);
			}
			return temp;
        }

		// NOTE the strange arrangement of bytes
        int GetInt(unsigned char* data, int startPosition)
        {
			int temp = (data[startPosition+2] << 24) | 
				       (data[startPosition+3] << 16) | 
					   (data[startPosition]   << 8) | 
					    data[startPosition+1];
			
			if (temp >= 0x80000000)
			{
				temp = temp ^ 0xFFFFFFFF;
				temp = -(temp + 1);
			}
			return temp;
        }

	};



}


