//----------------------------------------------------------------------------
//COPYRIGHT (C) 2005 SEGWAY INC.
//
//Segway provides this software "as is" with no warranty of any kind
//either expressed or implied, including, but not limited to the 
//implied warranties of merchantability, fitness for a particular 
//purpose, and noninfringement.
//
//Permission is granted to copy and modify this software for usage on 
//the RMP development platform.
//
//----------------------------------------------------------------------------
///	\file		usb_int.cpp
///	\brief		USB interface SW
///
///	- Author:	Eric L Pribyl
///	- Date:		May 2005
///
///Description:
///This module is the interface library to the ftd2xx.dll.
///The abstraction considers the USB as a CAN IO port.
///
///In its simplest form, an application only needs to
///1.  Establish a USB connection
///    usb_init()
///2.  Read and write as desired
///    usb_Read(...)  usb_Write(...)
///3.  Close the connection when done
///    usb_closedll()
///
///These functions are documented in the FTD2xx library specification.
///See http://www.ftdichip.com
///

#include <windows.h>
#include "ftd2xx.h"
#include "usb_int.h"

HMODULE	usb_hmodule;

FT_HANDLE usb_ftHandle;

int	usb_xmodem_sent = 0;
int usb_xmodem_send_length = 0;


PtrToWrite				usb_pWrite;
PtrToOpen				usb_pOpen; 
PtrToOpenEx				usb_pOpenEx; 
PtrToListDevices		usb_pListDevices; 
PtrToClose				usb_pClose;
PtrToRead				usb_pRead;
PtrToResetDevice		usb_pResetDevice;
PtrToPurge				usb_pPurge;
PtrToSetTimeouts		usb_pSetTimeouts;
PtrToGetQueueStatus		usb_pGetQueueStatus;
PtrToSetBaudRate		usb_pSetBaudRate;
PtrToSetLatencyTimer	usb_pSetLatencyTimer;


bool	usb_activep = 0;

//----------------------------------------------------------------------------
//			usb_init
//
// \brief	Initialize the USB interface
//
// \param	
//
// \return	int
//
// \remarks	This single call loads the library, and connects to the first
//			device found.
//
int usb_init()
{
	FT_STATUS status;
	DWORD num_devices;
	long	thisdevice = 0;
	
	//load the DLL
	usb_LoadDLL();


	status = usb_ListDevices(&num_devices, NULL, FT_LIST_NUMBER_ONLY);
	if (status != FT_OK
		|| num_devices <= 0)  {
		return -1;
	}


	status = usb_Open((PVOID) (thisdevice ));
	if (status == FT_OK)  {
		status = usb_SetBaudRate(FT_BAUD_460800);
		if (status != FT_OK)  {
			usb_Close();		//close open device
			return -2;
		}

		//set the latency timer to 1ms
		status = usb_SetLatencyTimer(1);
		if (status != FT_OK)  {
		//latency not set - but we won't know	return -82;
		}
		usb_ResetDevice();
		usb_Purge(FT_PURGE_RX | FT_PURGE_TX);
		usb_ResetDevice();
		usb_SetTimeouts(1000, 1000);	//extend timeout while board finishes reset
		Sleep(150);

	}

	usb_activep = 1;
	return 0;
}




//----------------------------------------------------------------------------
//			usb_active
//
// \brief	Predicate is the USB connection active?
//
// \param	
//
// \return	bool
//
// \remarks	
//
bool usb_active()
{
	return usb_activep;
}


//----------------------------------------------------------------------------
//			usb_closedll
//
// \brief	Close and unload the DLL
//
// \param	
//
// \return	int
//
// \remarks	
//
int usb_closedll()
{
	usb_activep = 0;
	return FreeLibrary(usb_hmodule);
}


//----------------------------------------------------------------------------
//			usb_LoadDLL
//
// \brief	Load the DLL and setup the library calls
//
// \param	
//
// \return	int
//
// \remarks	
//
int usb_LoadDLL()
{
	usb_hmodule = LoadLibrary("Ftd2xx.dll");
	if (usb_hmodule == NULL)  {
		return -1;
	}

	usb_pWrite = (PtrToWrite) GetProcAddress(usb_hmodule, "FT_Write");
	if (usb_pWrite == NULL)  {
		return -2;
	}

	usb_pRead = (PtrToRead) GetProcAddress(usb_hmodule, "FT_Read");
	if (usb_pRead == NULL) 	{
		return -3;
	}

	usb_pOpen = (PtrToOpen) GetProcAddress(usb_hmodule, "FT_Open");
	if (usb_pOpen == NULL) {
		return -4;
	}

	usb_pOpenEx = (PtrToOpenEx) GetProcAddress(usb_hmodule, "FT_OpenEx");
	if (usb_pOpenEx == NULL)  {
		return -5;
	}

	usb_pListDevices =
		(PtrToListDevices) GetProcAddress(usb_hmodule, "FT_ListDevices");
	if (usb_pListDevices == NULL)  {
		return -6;
	}

	usb_pClose = (PtrToClose) GetProcAddress(usb_hmodule, "FT_Close");
	if (usb_pClose == NULL) {
		return -7;
	}

	usb_pResetDevice =
		(PtrToResetDevice) GetProcAddress(usb_hmodule, "FT_ResetDevice");
	if (usb_pResetDevice == NULL)  {
		return -8;
	}

	usb_pPurge = (PtrToPurge) GetProcAddress(usb_hmodule, "FT_Purge");
	if (usb_pPurge == NULL)  {
		return -9;
	}

	usb_pSetTimeouts =
		(PtrToSetTimeouts) GetProcAddress(usb_hmodule, "FT_SetTimeouts");
	if (usb_pSetTimeouts == NULL)  {
		return -10;
	}

	usb_pGetQueueStatus =
		(PtrToGetQueueStatus) GetProcAddress(usb_hmodule, "FT_GetQueueStatus");
	if (usb_pGetQueueStatus == NULL)  {
		return -11;
	}
	usb_pSetBaudRate =
		(PtrToSetBaudRate) GetProcAddress(usb_hmodule, "FT_SetBaudRate");
	if (usb_pSetBaudRate == NULL)  {
		return -12;
	}

	usb_pSetBaudRate =
		(PtrToSetBaudRate) GetProcAddress(usb_hmodule, "FT_SetBaudRate");
	if (usb_pSetBaudRate == NULL)  {
		return -13;
	}

	usb_pSetLatencyTimer =
		(PtrToSetLatencyTimer) GetProcAddress(usb_hmodule, "FT_SetLatencyTimer");
	if (usb_pSetLatencyTimer == NULL)  {
		return -14;
	}
	return 0;
}

//----------------------------------------------------------------------------
//			usb_Write
//
// \brief	Write a buffer to the USB interface
//
// \param	LPVOID lpvBuffer
//
// \param	DWORD dwBuffSize
//
// \param	LPDWORD lpdwBytes
//
// \return	FT_STATUS
//
// \remarks	
//
FT_STATUS usb_Write(LPVOID lpvBuffer, DWORD dwBuffSize, LPDWORD lpdwBytes)
{
	if (!usb_pWrite)
	{
		return FT_INVALID_HANDLE;
	}

	return (*usb_pWrite) (usb_ftHandle, lpvBuffer, dwBuffSize, lpdwBytes);
}

//----------------------------------------------------------------------------
//			usb_Read
//
// \brief	Read into a buffer from the USB interface
//
// \param	LPVOID lpvBuffer
//
// \param	DWORD dwBuffSize
//
// \param	LPDWORD lpdwBytesRead
//
// \return	FT_STATUS
//
// \remarks	
//
FT_STATUS usb_Read(LPVOID lpvBuffer, DWORD dwBuffSize, LPDWORD lpdwBytesRead)
{
	if (!usb_pRead)
	{
		return FT_INVALID_HANDLE;
	}

	return (*usb_pRead) (usb_ftHandle, lpvBuffer, dwBuffSize, lpdwBytesRead);
}

//FT_STATUS usb_Read2(LPVOID lpvBuffer, DWORD dwBuffSize, LPDWORD lpdwBytesRead)
//{
//	if (!usb_pRead)
//	{
//		return FT_INVALID_HANDLE;
//	}
//	
//	DWORD queueLength;
//	DWORD ftstatus;
//
//	(*usb_pGetQueueStatus) (usb_ftHandle, &queueLength);
//	if (queueLength > 0)
//	{
//		ftstatus = (*usb_pRead) (usb_ftHandle, lpvBuffer, dwBuffSize, lpdwBytesRead);
//	}
//
//	//return (*usb_pRead) (usb_ftHandle, lpvBuffer, dwBuffSize, lpdwBytesRead);
//}


//----------------------------------------------------------------------------
//			usb_Open
//
// \brief	Open the interface to the USB
//
// \param	PVOID pvDevice
//
// \return	FT_STATUS
//
// \remarks
//
FT_STATUS usb_Open(PVOID pvDevice)
{
	if (!usb_pOpen)
	{
		return FT_INVALID_HANDLE;
	}

	return (*usb_pOpen) (pvDevice, &usb_ftHandle);
}

//----------------------------------------------------------------------------
//			usb_OpenEx
//
// \brief	
//
// \param	PVOID pArg1
//
// \param	DWORD dwFlags
//
// \return	FT_STATUS
//
// \remarks	
//
FT_STATUS usb_OpenEx(PVOID pArg1, DWORD dwFlags)
{
	if (!usb_pOpenEx)
	{
		return FT_INVALID_HANDLE;
	}

	return (*usb_pOpenEx) (pArg1, dwFlags, &usb_ftHandle);
}


//----------------------------------------------------------------------------
//			usb_ListDevices
//
// \brief	List the available devices
//
// \param	PVOID pArg1
//
// \param	PVOID pArg2
//
// \param	DWORD dwFlags
//
// \return	FT_STATUS
//
// \remarks	
//
FT_STATUS usb_ListDevices(PVOID pArg1, PVOID pArg2, DWORD dwFlags)
{
	if (!usb_pListDevices)
	{
		return FT_INVALID_HANDLE;
	}

	return (*usb_pListDevices) (pArg1, pArg2, dwFlags);
}


//----------------------------------------------------------------------------
//			usb_Close
//
// \brief	Close the connection
//
// \param	
//
// \return	FT_STATUS
//
// \remarks	
//
FT_STATUS usb_Close()
{
	if (!usb_pClose)
	{
		return FT_INVALID_HANDLE;
	}
	return (*usb_pClose) (usb_ftHandle);
}




//----------------------------------------------------------------------------
//			usb_ResetDevice
//
// \brief	Reset the FTD device
//
// \param	
//
// \return	FT_STATUS
//
// \remarks	
//
FT_STATUS usb_ResetDevice()
{
	if (!usb_pResetDevice)
	{
		return FT_INVALID_HANDLE;
	}

	return (*usb_pResetDevice) (usb_ftHandle);
}



//----------------------------------------------------------------------------
//			usb_Purge
//
// \brief	Purge pending TX and RX
//
// \param	ULONG dwMask
//
// \return	FT_STATUS
//
// \remarks	
//
FT_STATUS usb_Purge(ULONG dwMask)
{
	if (!usb_pPurge)
	{
		return FT_INVALID_HANDLE;
	}

	return (*usb_pPurge) (usb_ftHandle, dwMask);
}



//----------------------------------------------------------------------------
//			usb_SetTimeouts
//
// \brief	Set timeouts for read and write operations
//
// \param	ULONG dwReadTimeout
//
// \param	ULONG dwWriteTimeout
//
// \return	FT_STATUS
//
// \remarks	
//
FT_STATUS usb_SetTimeouts(ULONG dwReadTimeout, ULONG dwWriteTimeout)
{
	if (!usb_pSetTimeouts)
	{
		return FT_INVALID_HANDLE;
	}

	return (*usb_pSetTimeouts) (usb_ftHandle, dwReadTimeout, dwWriteTimeout);
}


//----------------------------------------------------------------------------
//			usb_GetQueueStatus
//
// \brief	How big the the RX queue?
//
// \param	LPDWORD lpdwAmountInRxQueue
//
// \return	FT_STATUS
//
// \remarks	
//
FT_STATUS usb_GetQueueStatus(LPDWORD lpdwAmountInRxQueue)
{
	if (!usb_pGetQueueStatus)
	{
		return FT_INVALID_HANDLE;
	}

	if (!usb_active())
		return FT_INVALID_HANDLE;

	return (*usb_pGetQueueStatus) (usb_ftHandle, lpdwAmountInRxQueue);
}

//----------------------------------------------------------------------------
//			usb_SetBaudRate
//
// \brief	Set baud rate
//
// \param	ULONG ulDesiredBaud
//
// \return	FT_STATUS
//
// \remarks	
//
FT_STATUS usb_SetBaudRate(ULONG ulDesiredBaud)
{
	if (!usb_pSetBaudRate)
	{
		return FT_INVALID_HANDLE;
	}
	return (*usb_pSetBaudRate) (usb_ftHandle, ulDesiredBaud);
}

//----------------------------------------------------------------------------
//			usb_SetLatencyTimer
//
// \brief	Set latency timer
//
// \param	UCHAR ucDesiredLatency
//
// \return	FT_STATUS
//
// \remarks	
//
FT_STATUS usb_SetLatencyTimer(UCHAR ucDesiredLatency)
{
	if (!usb_pSetLatencyTimer)
	{
		return FT_INVALID_HANDLE;
	}
	return (*usb_pSetLatencyTimer) (usb_ftHandle, ucDesiredLatency);
}




//----------------------------------------------------------------------------
//			usb_message_format
//
// \brief	Format a message for the USB
//
// \param	unsigned char *sbytes
//
// \return	int
//
// \remarks	pre-pend the header and caluate the checksum
//			for the message being sent.  Note that the buffer
//			is expected to be exactly 18 bytes in length.
//
int usb_message_format(unsigned char *sbytes)
{
	sbytes[0] = 0xf0;	//SOH


	unsigned short checksum;
	unsigned short checksum_hi;
	checksum = 0;
	for(int i = 0; i < 17; i++)
	{
		checksum += (short)sbytes[i];
	}
	checksum_hi = (unsigned short)(checksum >> 8);
	checksum &= 0xff;
	checksum += checksum_hi;
	checksum_hi = (unsigned short)(checksum >> 8);
	checksum &= 0xff;
	checksum += checksum_hi;
	checksum = (~checksum + 1) & 0xff;
	sbytes[17] = (unsigned char)checksum;
	return 0;
}



//----------------------------------------------------------------------------
//			usb_can_send
//
// \brief	Send a CAN message via USB
//
// \param	int channel
//
// \param	int id
//
// \param	unsigned short *data
//
// \return	int
//
// \remarks	
//
int usb_can_send(int channel, int id, unsigned short *data)
{
	unsigned long ret_bytes;
	unsigned char sbytes[18];

	sbytes[1] = 0x55;
	sbytes[2] = 0;	//CAN A
	sbytes[3] = 0;
	sbytes[4] = 0;
	sbytes[5] = 0;
	sbytes[6] = id >> 8;
	sbytes[7] = id ;
	sbytes[8] = 0;
	sbytes[9] =  data[0] >> 8;
	sbytes[10] = data[0] & 0x0ff;
	sbytes[11] = data[1] >> 8;
	sbytes[12] = data[1] & 0x0ff;
	sbytes[13] = data[2] >> 8;
	sbytes[14] = data[2] & 0x0ff;
	sbytes[15] = data[3] >> 8;
	sbytes[16] = data[3] & 0x0ff;
	sbytes[17] = 0;

	usb_message_format(sbytes);
	if (channel & 1)  {
		usb_Write(sbytes, 18, &ret_bytes);
	}

	//mod the message to broadcast on CANB
	sbytes[2] = 1;

	usb_message_format(sbytes);
	if (channel & 2)  {
		usb_Write(sbytes, 18, &ret_bytes);
	}
	return 0;
}
	



#define	HIGHBYTE(data)	(char)(data >> 8)
#define	LOWBYTE(data)	(char)data
//XMODEM protocol constants
#define NAK			0x15		// negative acknowledgment
#define ACK			0x06		// acknowledgment
#define SOH			0x01		// start of header
#define EOT			0x04		// end of text
#define CAN			0x18		// cancel packet


unsigned short icrctb[] = {
		0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50A5, 0x60C6, 0x70E7,
		0x8108, 0x9129, 0xA14A, 0xB16B, 0xC18C, 0xD1AD, 0xE1CE, 0xF1EF,
		0x1231, 0x0210, 0x3273, 0x2252, 0x52B5, 0x4294, 0x72F7, 0x62D6,
		0x9339, 0x8318, 0xB37B, 0xA35A, 0xD3BD, 0xC39C, 0xF3FF, 0xE3DE,
		0x2462, 0x3443, 0x0420, 0x1401, 0x64E6, 0x74C7, 0x44A4, 0x5485,
		0xA56A, 0xB54B, 0x8528, 0x9509, 0xE5EE, 0xF5CF, 0xC5AC, 0xD58D,
		0x3653, 0x2672, 0x1611, 0x0630, 0x76D7, 0x66F6, 0x5695, 0x46B4,
		0xB75B, 0xA77A, 0x9719, 0x8738, 0xF7DF, 0xE7FE, 0xD79D, 0xC7BC,
		0x48C4, 0x58E5, 0x6886, 0x78A7, 0x0840, 0x1861, 0x2802, 0x3823,
		0xC9CC, 0xD9ED, 0xE98E, 0xF9AF, 0x8948, 0x9969, 0xA90A, 0xB92B,
		0x5AF5, 0x4AD4, 0x7AB7, 0x6A96, 0x1A71, 0x0A50, 0x3A33, 0x2A12,
		0xDBFD, 0xCBDC, 0xFBBF, 0xEB9E, 0x9B79, 0x8B58, 0xBB3B, 0xAB1A,
		0x6CA6, 0x7C87, 0x4CE4, 0x5CC5, 0x2C22, 0x3C03, 0x0C60, 0x1C41,
		0xEDAE, 0xFD8F, 0xCDEC, 0xDDCD, 0xAD2A, 0xBD0B, 0x8D68, 0x9D49,
		0x7E97, 0x6EB6, 0x5ED5, 0x4EF4, 0x3E13, 0x2E32, 0x1E51, 0x0E70,
		0xFF9F, 0xEFBE, 0xDFDD, 0xCFFC, 0xBF1B, 0xAF3A, 0x9F59, 0x8F78,
		0x9188, 0x81A9, 0xB1CA, 0xA1EB, 0xD10C, 0xC12D, 0xF14E, 0xE16F,
		0x1080, 0x00A1, 0x30C2, 0x20E3, 0x5004, 0x4025, 0x7046, 0x6067,
		0x83B9, 0x9398, 0xA3FB, 0xB3DA, 0xC33D, 0xD31C, 0xE37F, 0xF35E,
		0x02B1, 0x1290, 0x22F3, 0x32D2, 0x4235, 0x5214, 0x6277, 0x7256,
		0xB5EA, 0xA5CB, 0x95A8, 0x8589, 0xF56E, 0xE54F, 0xD52C, 0xC50D,
		0x34E2, 0x24C3, 0x14A0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
		0xA7DB, 0xB7FA, 0x8799, 0x97B8, 0xE75F, 0xF77E, 0xC71D, 0xD73C,
		0x26D3, 0x36F2, 0x0691, 0x16B0, 0x6657, 0x7676, 0x4615, 0x5634,
		0xD94C, 0xC96D, 0xF90E, 0xE92F, 0x99C8, 0x89E9, 0xB98A, 0xA9AB,
		0x5844, 0x4865, 0x7806, 0x6827, 0x18C0, 0x08E1, 0x3882, 0x28A3,
		0xCB7D, 0xDB5C, 0xEB3F, 0xFB1E, 0x8BF9, 0x9BD8, 0xABBB, 0xBB9A,
		0x4A75, 0x5A54, 0x6A37, 0x7A16, 0x0AF1, 0x1AD0, 0x2AB3, 0x3A92,
		0xFD2E, 0xED0F, 0xDD6C, 0xCD4D, 0xBDAA, 0xAD8B, 0x9DE8, 0x8DC9,
		0x7C26, 0x6C07, 0x5C64, 0x4C45, 0x3CA2, 0x2C83, 0x1CE0, 0x0CC1,
		0xEF1F, 0xFF3E, 0xCF5D, 0xDF7C, 0xAF9B, 0xBFBA, 0x8FD9, 0x9FF8,
		0x6E17, 0x7E36, 0x4E55, 0x5E74, 0x2E93, 0x3EB2, 0x0ED1, 0x1EF0
	};


//----------------------------------------------------------------------------
//			int usb_XModem_Send
//
// \brief	Xmodem to send a buffer
//
// \param	unsigned char *buffer
//
// \param	unsigned int send_len
//
// \return	unsigned
//
// \remarks	
//
unsigned int usb_XModem_Send(unsigned char *buffer, unsigned int send_len)
{
	
	unsigned int byte_count;	// total of bytes downloaded
	unsigned int i;				// data counter
	unsigned char data;			// data buffer
	unsigned char s[4];			// generic buffer
	unsigned char tx_data[128];	// tx buffer
	unsigned char packet_num;	//  packet number
	unsigned int packet_total;	//  packet total
	unsigned int packets_sent;
	unsigned short cword;		// CRC checking word
	unsigned long ret_bytes;

	//clear counters
	byte_count = 0;
	packet_num = 0;
	packet_total = send_len / 128;
	packets_sent = 0;

	usb_xmodem_send_length = send_len;

	// look for a "C" 
	for(i = 0; i < 10; i++)
	{
		usb_Read(&data, 1, &ret_bytes);
		if (data == 'C')
			break;
	}

	// check why it left previous loop
	if (data != 'C')			//if unsuccessful return
	{
		s[0] = CAN;
		s[1] = CAN;
		usb_Write(s, 2, &ret_bytes);
		return (0);
	}

	// send SOH
	usb_Purge(FT_PURGE_RX || FT_PURGE_TX);
	s[0] = SOH;
	usb_Write(s, 1, &ret_bytes);


	while (1)
	{
		/*m_PacketsSent.Format("%d", packets_sent++);
		m_PacketsRemaining.Format("%d", packet_total--);
		UpdateData(FALSE);
		UpdateWindow();
		*/
		Sleep(20);

		// send packet number and packet number compliment
		packet_num++;
		s[0] = packet_num;
		s[1] = ~packet_num;
		usb_Write(s, 2, &ret_bytes);

		//clear crc
		cword = 0;

		//send 128 bytes of data
		for(i = 0; i < 128; i++, byte_count++, buffer++)
		{
			// send stored data
			tx_data[i] = *buffer;
			//calculate crc
			cword =
				(icrctb[(unsigned char)(*buffer ^ HIGHBYTE(cword))]) ^
				((unsigned short)LOWBYTE(cword) << 8);
		}
		//Purge(FT_PURGE_RX || FT_PURGE_TX);
		usb_Write(tx_data, 64, &ret_bytes);
		usb_Write(&tx_data[64], 64, &ret_bytes);

		// send crc
		s[0] = HIGHBYTE(cword);
		s[1] = LOWBYTE(cword);
		usb_Write(s, 2, &ret_bytes);

		// get the response from the remote computer
		do
		{
			usb_Read(&data, 1, &ret_bytes);
		} while (ret_bytes < 1);

		//check for a proper reply
		if ((data != ACK) && (data != NAK) && (data != CAN))
		{
			s[0] = CAN;
			s[1] = CAN;
			usb_Write(s, 2, &ret_bytes);
			return (0);
		}

		// check for error in transfer
		if (data == NAK)
		{
			buffer -= 128;
			byte_count -= 128;	//recount bytes
			packet_num--;		//decrement packet count
		}

		//check for EOT
		if (byte_count >= (send_len - 1))
		{
			s[0] = EOT;
			usb_Write(s, 1, &ret_bytes);
			break;
		}

		usb_xmodem_sent = byte_count;

		// send SOH
		s[0] = SOH;
		usb_Write(s, 1, &ret_bytes);

	}

	//check for cancel
	if (data == CAN)
	{
		return (0);
	}
	//else transfer successful
	else
	{
	}

	
	usb_xmodem_send_length = 0;

	return 1;
}


//----------------------------------------------------------------------------
//			usb_reset
//
// \brief	Send reset command to USB
//
// \param	
//
// \return	int
//
// \remarks	
//
int usb_reset()
{
	unsigned char sbytes[18];
	unsigned long ret_bytes;

	memset(&sbytes[0], 0, sizeof(sbytes));
	sbytes[1] = 0x55;	//header
	sbytes[2] = 5;		//command

	usb_message_format(sbytes);
	usb_Write(sbytes, 18, &ret_bytes);
	return 1;
}
