//------------------------------------------------------------------------------
// SegwayBattery.cs
//
// Author: Ben Axelrod
// Date: April 2007
// Version 1.0
//
//------------------------------------------------------------------------------
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using System;
using System.Collections.Generic;

using submgr = Microsoft.Dss.Services.SubscriptionManager;
using segway = Robotics.SegwayBase.Proxy;
using battery = Microsoft.Robotics.Services.Battery.Proxy;
using speech = Microsoft.Robotics.Technologies.Speech.TextToSpeech.Proxy;
using W3C.Soap;

// Powerbase Battery Nominal Voltage  72 Volts
// Powerbase Battery Voltage          4 counts/Volt
// User Interface Battery Voltage     1.4 + counts * 0.0125

namespace Robotics.SegwayBattery
{
    /// <summary>
    /// SegwayBattery Contract
    /// </summary>
    public sealed class Contract
    {
        /// The Unique Contract Identifier for the SegwayBattery service
        public const String Identifier = "http://schemas.tempuri.org/2007/04/segwaybattery.html";
    }

    /// <summary>
    /// Segway Battery Service
    /// </summary>
    [Contract(Contract.Identifier)]
    [AlternateContract("http://schemas.microsoft.com/2006/06/battery.html")]
    public class SegwayBatteryService : DsspServiceBase
    {
        /// <summary>
        /// _state
        /// </summary>
        private battery.BatteryState _state = new battery.BatteryState();
        private battery.BatteryState _stateUI = new battery.BatteryState();

        /// <summary>
        /// _main Port
        /// </summary>
        [ServicePort("/segwaybattery", AllowMultipleInstances=true)]
        private battery.BatteryOperations _mainPort = new battery.BatteryOperations();

        [Partner("Segway", Contract = segway.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate, Optional = false)]
        private segway.SegwayBaseOperations _segwayPort = new segway.SegwayBaseOperations();

        [Partner("speech", Contract = speech.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate)]
        speech.SpeechTextOperations _speechPort = new speech.SpeechTextOperations();

        [Partner("SubMgr", Contract = submgr.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.CreateAlways, Optional = false)]
        private submgr.SubscriptionManagerPort _subMgrPort = new submgr.SubscriptionManagerPort();


        /// <summary>
        /// Default Service Constructor
        /// </summary>
        public SegwayBatteryService(DsspServiceCreationPort creationPort) : base(creationPort)
        {
        }

        Interleave _mainInterleave;
        private bool _subscribed = false;

        /// <summary>
        /// Service Start
        /// </summary>
        protected override void Start()
        {
            #region base.Start()
            // Listen for each of my handlers
            _mainInterleave = ActivateDsspOperationHandlers();

            // Publish the service to the local Node Directory
            DirectoryInsert();

            // display HTTP service Uri
            LogInfo(LogGroups.Console, "Service uri: ");
            #endregion
			
            //init state
            _state.MaxBatteryPower = 288; // = (72 volts) * (4 counts / volt)
            _state.PercentBatteryPower = 1.0;
            _state.PercentCriticalBattery = 0.2;

            _stateUI.MaxBatteryPower = 608; // = ( (9 volts) - (1.4 volts) ) / ( 0.0125 volts / count)
            _stateUI.PercentBatteryPower = 1.0;
            _stateUI.PercentCriticalBattery = 0;

            SubscribeToSegway();

        }

        /// <summary>
        /// Subscribe to appropriate data message update on segway
        /// </summary>
        private void SubscribeToSegway()
        {
            // Create a notification port
            segway.SegwayBaseOperations _notificationPort = new segway.SegwayBaseOperations();

            //create a custom subscription request
            segway.MySubscribeRequestType request = new segway.MySubscribeRequestType();

            //select appropriate data message
            request.DataMessageNumbers = new List<int>(1);
            request.DataMessageNumbers.Add(6); //WARNING: Magic number

            //Subscribe to the Segway and wait for a response
            Activate(
                Arbiter.Choice(_segwayPort.SelectiveSubscribe(request, _notificationPort),
                    delegate(SubscribeResponseType Rsp)
                    {
                        //update our state with subscription status
                        _subscribed = true;

                        LogInfo("SegwayEncoderLeftRight subscription success");

                        //Subscription was successful, start listening for sensor change notifications
                        _mainInterleave.CombineWith(new Interleave(
                            new ExclusiveReceiverGroup(
                                Arbiter.Receive<segway.Replace>(true, _notificationPort, SensorNotificationHandler)
                            ),
                            new ConcurrentReceiverGroup()
                        ));
                    },
                    delegate(Fault F)
                    {
                        LogError("SegwayEncoderLeftRight subscription failed");
                    }
                )
            );
        }

        /// <summary>
        /// Handle sensor update message from Segway
        /// </summary>
        public void SensorNotificationHandler(segway.Replace notify)
        {
            _state.PercentBatteryPower = (double)notify.Body.PBBatteryVoltage / _state.MaxBatteryPower;
            _stateUI.PercentBatteryPower = (double)notify.Body.UIBatteryVoltage / _stateUI.MaxBatteryPower;

            if (_state.PercentBatteryPower < _state.PercentCriticalBattery)
            {
                speech.SayTextSynchRequest sayme = new speech.SayTextSynchRequest();
                sayme.SpeechText = "Warning!  Power base battery voltage at " + Math.Round(_state.PercentBatteryPower * 100.0, 1) + " percent!";
                _speechPort.SayTextSynch(sayme);
                
                _state.PercentCriticalBattery /= 2.0; //next warning a little later

            }

            //notify subscribers on any state change
            base.SendNotification(_subMgrPort, new battery.Replace(_state));
        }

        /// <summary>
        /// Get Handler
        /// </summary>
        /// <param name="get"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> GetHandler(battery.Get get)
        {
            get.ResponsePort.Post(_state);

            speech.SayTextSynchRequest sayme = new speech.SayTextSynchRequest();
            sayme.SpeechText = "Power base battery voltage is " + Math.Round(_state.PercentBatteryPower * 72.0,1) + " volts.  " +
                               "User interface battery voltage is " + Math.Round(_stateUI.PercentBatteryPower * 9.0, 1) + " volts.";

            _speechPort.SayTextSynch(sayme);

            yield break;
        }

        /// <summary>
        /// Replace Handler
        /// </summary>
        /// <param name="replace"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> ReplaceHandler(battery.Replace replace)
        {
            _state = replace.Body;
            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);
            yield break;
        }


        /// <summary>
        /// Subscribe Handler
        /// </summary>
        /// <param name="subscribe"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> SubscribeHandler(battery.Subscribe subscribe)
        {
            yield return Arbiter.Choice(
                SubscribeHelper(_subMgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    SendNotification<battery.Replace>(_subMgrPort, subscribe.Body.Subscriber, new battery.Replace(_state));
                },
                delegate(Exception fault)
                {
                    LogError(fault);
                }
            );
        }


        /// <summary>
        /// SetCriticalLevel Handler
        /// </summary>
        /// <param name="update"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> SetCriticalLevelHandler(battery.SetCriticalLevel update)
        {
            _state.PercentCriticalBattery = update.Body.PercentCriticalBattery;
            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

    }
}
