//------------------------------------------------------------------------------
// SegwayDrive.cs
//
// Author: Ben Axelrod
// Date: April 2007
// Version 1.0
//
//------------------------------------------------------------------------------
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.Core.DsspHttp;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using Microsoft.Dss.Services.SubscriptionManager;
using Microsoft.Dss.Core.DsspHttpUtilities;
using System;
using System.Net;
using System.ComponentModel;
using System.Collections.Generic;
using System.Collections.Specialized;
using W3C.Soap;
using cons = Microsoft.Dss.Services.Constructor;


using drive = Microsoft.Robotics.Services.Drive.Proxy;
using motor = Microsoft.Robotics.Services.Motor.Proxy;
using encoder = Microsoft.Robotics.Services.Encoder.Proxy;

using segway = Robotics.SegwayBase.Proxy;
using foreaftencoder = Robotics.SegwayEncoder.ForeAftYaw;
using leftrightencoder = Robotics.SegwayEncoder.LeftRight;






namespace Robotics.SegwayDrive
{

    /// <summary>
    /// SegwayDrive Contract
    /// </summary>
    public sealed class Contract
    {
        /// The Unique Contract Identifier for the SegwayDrive service
        public const String Identifier = "http://schemas.tempuri.org/2007/03/segwaydrive.html";
    }


    /// <summary>
    /// Segway Services Service
    /// </summary>
    [Contract(Contract.Identifier)]
    [AlternateContract("http://schemas.microsoft.com/robotics/2006/05/drive.html")]
    public class SegwayDriveService : DsspServiceBase
    {
        /// <summary>
        /// Microsoft.Robotics.Services.Drive.SegwayDrive Partner
        /// </summary>
        [Partner("SubMgr", Contract="http://schemas.microsoft.com/xw/2005/01/subscriptionmanager.html", CreationPolicy=PartnerCreationPolicy.CreateAlways, Optional=false)]
        private SubscriptionManagerPort _subMgrPort = new SubscriptionManagerPort();
        
        /// <summary>
        /// _main Port
        /// </summary>
        [ServicePort("/segwaydrive", AllowMultipleInstances=false)]
        private drive.DriveOperations _mainPort = new drive.DriveOperations();

        [Partner("Segway", Contract = segway.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate, Optional = false)]
        private segway.SegwayBaseOperations _segwayPort = new segway.SegwayBaseOperations();

        [Partner("ForeAftEncoder", Optional = true, Contract = foreaftencoder.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExisting)]
        foreaftencoder.SegwayEncoderForeAftYawOperations _foreaftEncoderPort = new foreaftencoder.SegwayEncoderForeAftYawOperations();
        //foreaftencoder.SegwayEncoderForeAftYawOperations _foreaftEncoderNotify = new foreaftencoder.SegwayEncoderForeAftYawOperations();

        [Partner("LeftRightEncoder", Optional = true, Contract = leftrightencoder.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExisting)]
        leftrightencoder.SegwayEncoderLeftRightOperations _leftrightEncoderPort = new leftrightencoder.SegwayEncoderLeftRightOperations();
        //leftrightencoder.SegwayEncoderLeftRightOperations _leftrightEncoderNotify = new leftrightencoder.SegwayEncoderLeftRightOperations();

        /// <summary>
        /// Robotics.SegwayDrive.SegwayDriveService State
        /// </summary>
        //[InitialStatePartner(Optional=true, ServiceUri="Robotics.SegwayDrive.SegwayDriveService.Config.xml")]
        private drive.DriveDifferentialTwoWheelState _state = new drive.DriveDifferentialTwoWheelState();

        /// <summary>
        /// XSLT
        /// </summary>
        [EmbeddedResource("Robotics.SegwayServices.SegwayDrive.xslt")]
        string _transform = null;

        private DsspHttpUtilitiesPort _httpUtilities = new DsspHttpUtilitiesPort();


        /// <summary>
        /// Default Service Constructor
        /// </summary>
        public SegwayDriveService(DsspServiceCreationPort creationPort) : 
                base(creationPort)
        {
        }

        private readonly double VelSafteyScale = 0.125;
        private readonly double TurnSafteyScale = 0.5;
        private readonly int MaxVelocity = 1176;
        private readonly int MaxTurn = 1024;

        private enum DriveStates
        {
            Stopped,
            DriveDistance,
            DrivePower,            
            RotateDegrees,
            DriveSpeed
        }

        /// <summary>
        /// additional state
        /// </summary>
        private DriveStates _driveState;

        /// <summary>
        /// Service Start
        /// </summary>
        protected override void Start()
        {
			base.Start();
			// Add service specific initialization here.

            //for HttpPost
            _httpUtilities = DsspHttpUtilitiesService.Create(Environment);

            //for XSLT and images
            MountEmbeddedResources("/SegwayServices");

            _state.IsEnabled = true;
            _state.DistanceBetweenWheels = 0.54;

            _state.LeftWheel = new motor.WheeledMotorState();
            _state.LeftWheel.Radius = 0.24;
            _state.LeftWheel.EncoderState = new encoder.EncoderState();
            _state.LeftWheel.MotorState = new motor.MotorState();
            
            _state.RightWheel = new motor.WheeledMotorState();
            _state.RightWheel.Radius = 0.24;
            _state.RightWheel.EncoderState = new encoder.EncoderState();
            _state.RightWheel.MotorState = new motor.MotorState();

            _driveState = DriveStates.Stopped;

        }

        /// <summary>
        /// Get Handler
        /// </summary>
        /// <param name="get"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> GetHandler(drive.Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        /// <summary>
        /// Update Handler
        /// </summary>
        /// <param name="update"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> UpdateHandler(drive.Update update)
        {
            // TODO: Implement Update operations here.
            throw new NotImplementedException("TODO: Implement Update operations here.");
        }

        /// <summary>
        /// Subscribe Handler
        /// </summary>
        /// <param name="subscribe"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> SubscribeHandler(drive.Subscribe subscribe)
        {
            base.SubscribeHelper(_subMgrPort, subscribe.Body, subscribe.ResponsePort);
            yield break;
        }

        /// <summary>
        /// ReliableSubscribe Handler
        /// </summary>
        /// <param name="subscribe"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> ReliableSubscribeHandler(drive.ReliableSubscribe subscribe)
        {
            base.SubscribeHelper(_subMgrPort, subscribe.Body, subscribe.ResponsePort);
            yield break;
        }

        /// <summary>
        /// EnableDrive Handler
        /// </summary>
        /// <param name="update"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> EnableDriveHandler(drive.EnableDrive update)
        {
            _state.IsEnabled = update.Body.Enable;
            _state.TimeStamp = DateTime.Now;
            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

        /// <summary>
        /// SetDrivePower Handler
        /// </summary>
        /// <param name="update"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> SetDrivePowerHandler(drive.SetDrivePower update)
        {
            if (_state.IsEnabled)
            {
                //Console.WriteLine("in SetDrivePowerHandler power " + update.Body.LeftWheelPower + " with power " + update.Body.RightWheelPower);

                _driveState = DriveStates.DrivePower;
                _state.LeftWheel.MotorState.CurrentPower = update.Body.LeftWheelPower;
                _state.RightWheel.MotorState.CurrentPower = update.Body.RightWheelPower;
                _state.TimeStamp = DateTime.Now;

                int velocity = (int)Math.Round(VelSafteyScale * MaxVelocity * (_state.LeftWheel.MotorState.CurrentPower + _state.RightWheel.MotorState.CurrentPower) / 2.0);
                int turn = (int)Math.Round(TurnSafteyScale * MaxTurn * (_state.RightWheel.MotorState.CurrentPower - _state.LeftWheel.MotorState.CurrentPower) / 2.0);

                segway.DriveMessage drv = new segway.DriveMessage();
                drv.Velocity = velocity;
                drv.Turn = turn;
                _segwayPort.Drive(drv);
            }

            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

        /// <summary>
        /// SetDriveSpeed Handler
        /// </summary>
        /// <param name="update"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> SetDriveSpeedHandler(drive.SetDriveSpeed update)
        {
            // TODO: Implement Update operations here.
            throw new NotImplementedException("TODO: Implement Update operations here.");
        }


        #region Rotate Degrees

        /// <summary>
        /// RotateDegrees Handler
        /// </summary>
        /// <param name="update"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> RotateDegreesHandler(drive.RotateDegrees update)
        {
            ValidateDriveConfiguration(true);
            //Console.WriteLine("in RotateDegreesHandler rotating " + update.Body.Degrees + " with power " + update.Body.Power);
            _state.TimeStamp = DateTime.Now;
            _driveState = DriveStates.RotateDegrees;
            SpawnIterator<double, double, PortSet<DefaultUpdateResponseType, Fault>>(update.Body.Degrees, update.Body.Power, update.ResponsePort, RotateUntilDegrees);

            //response gets posted when finished motion, or behavior overriden
            //update.ResponsePort.Post(new DefaultUpdateResponseType());

            yield break;
        }

        /// <summary>
        /// drives a specified number of meters
        /// </summary>
        IEnumerator<ITask> RotateUntilDegrees(double degrees, double power, PortSet<DefaultUpdateResponseType, Fault> responseport)
        {
            foreaftencoder.SegwayEncoderForeAftYawState startstate = null;

            //log state of encoders
            foreaftencoder.Get getmsg = new foreaftencoder.Get();
            _foreaftEncoderPort.Post(getmsg);
            yield return (Arbiter.Choice(getmsg.ResponsePort,
                delegate(foreaftencoder.SegwayEncoderForeAftYawState response)
                {
                    startstate = response;
                },
                delegate(Fault fault) { LogError(fault); }
            ));

            //compute tick to stop at
            int stopat = (int)Math.Round(startstate.CurrentYawReading + degrees * (double)startstate.TicksPerRevolution / 360.0);

            double Pow;
            if (degrees > 0)
                Pow = power;
            else
                Pow = -power;

            // Subscribe to the encoder on internal port
            foreaftencoder.SegwayEncoderForeAftYawOperations _foreaftNotificationPort = new foreaftencoder.SegwayEncoderForeAftYawOperations();

            foreaftencoder.Subscribe op = new foreaftencoder.Subscribe();
            op.Body = new SubscribeRequestType();
            op.NotificationPort = _foreaftNotificationPort;
            _foreaftEncoderPort.Post(op);
            yield return (Arbiter.Choice(op.ResponsePort,
                delegate(SubscribeResponseType response)
                {
                    //subscription was successful, start listening for encoder replace messages
                    Activate(Arbiter.Receive<foreaftencoder.Replace>(false, _foreaftNotificationPort,
                        delegate(foreaftencoder.Replace replace)
                        {
                            MotorStopper(DriveStates.RotateDegrees, _foreaftNotificationPort, replace, stopat, Pow, responseport);
                        }));
                },
                delegate(Fault fault) { LogError(fault); }
            ));

            //start moving

            _state.LeftWheel.MotorState.CurrentPower = -Pow;
            _state.RightWheel.MotorState.CurrentPower = Pow;
            int velocity = 0;
            int turn = (int)Math.Round(TurnSafteyScale * MaxTurn * Pow);

            segway.DriveMessage drv = new segway.DriveMessage();
            drv.Velocity = velocity;
            drv.Turn = turn;
            _segwayPort.Drive(drv);
        }


        #endregion
        


        #region Drive Distance

        /// <summary>
        /// DriveDistance Handler
        /// </summary>
        /// <param name="update"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> DriveDistanceHandler(drive.DriveDistance update)
        {
            ValidateDriveConfiguration(true);
            _driveState = DriveStates.DriveDistance;
            _state.TimeStamp = DateTime.Now;

            SpawnIterator<double, double, PortSet<DefaultUpdateResponseType, Fault>>(update.Body.Distance, update.Body.Power, update.ResponsePort, DriveUntilDistance);

            //response gets posted when finished motion, or behavior overriden
            //update.ResponsePort.Post(new DefaultUpdateResponseType());
            yield break;
        }

        /// <summary>
        /// drives a specified number of meters
        /// </summary>
        IEnumerator<ITask> DriveUntilDistance(double distance, double power, PortSet<DefaultUpdateResponseType, Fault> responseport)
        {
            foreaftencoder.SegwayEncoderForeAftYawState startstate = null;
            
            //log state of encoders
            foreaftencoder.Get getmsg = new foreaftencoder.Get();
            _foreaftEncoderPort.Post(getmsg);
            yield return (Arbiter.Choice(getmsg.ResponsePort,
                delegate(foreaftencoder.SegwayEncoderForeAftYawState response)
                {
                    startstate = response;
                },
                delegate(Fault fault) { LogError(fault); }
            ));

            //compute ticks to stop at
            int stopat = (int)Math.Round(startstate.CurrentForeAftReading + distance * startstate.TicksPerMeter);

            double Pow;
            if (distance > 0)
                Pow = power;
            else
                Pow = -power;

            // Subscribe to the encoder on internal port
            foreaftencoder.SegwayEncoderForeAftYawOperations _foreaftNotificationPort = new foreaftencoder.SegwayEncoderForeAftYawOperations();

            foreaftencoder.Subscribe op = new foreaftencoder.Subscribe();
            op.Body = new SubscribeRequestType();
            op.NotificationPort = _foreaftNotificationPort;
            _foreaftEncoderPort.Post(op);
            yield return (Arbiter.Choice(op.ResponsePort,
                delegate(SubscribeResponseType response)
                {
                    //subscription was successful, start listening for encoder replace messages
                    Activate(Arbiter.Receive<foreaftencoder.Replace>(false, _foreaftNotificationPort,
                        delegate(foreaftencoder.Replace replace)
                        {
                            MotorStopper(DriveStates.DriveDistance, _foreaftNotificationPort, replace, stopat, Pow, responseport);
                        }));
                },
                delegate(Fault fault) { LogError(fault); }
            ));

            //start moving
            _state.LeftWheel.MotorState.CurrentPower = Pow;
            _state.RightWheel.MotorState.CurrentPower = Pow;
            int velocity = (int)Math.Round(VelSafteyScale * MaxVelocity * Pow);
            int turn = 0;

            segway.DriveMessage drv = new segway.DriveMessage();
            drv.Velocity = velocity;
            drv.Turn = turn;
            _segwayPort.Drive(drv);
        }

        #endregion

        /// <summary>
        /// monitors the encoders and stops the motors when nessisary
        /// aborts if some other command changes drive state
        /// </summary>
        void MotorStopper(DriveStates state, //either DriveStates.DriveDistance or DriveStates.RotateDegrees.  so we know if this operation has been over-ridden
                          foreaftencoder.SegwayEncoderForeAftYawOperations encoderNotificationPort, //port to listen for encoder updates on
                          foreaftencoder.Replace msg, //the actual encoder update message
                          int stopAt, //encoder ticks to stop at
                          double power, //which way the motors were set. so we know if goal is infront or behind us
                          PortSet<DefaultUpdateResponseType, Fault> responseport) //port to send reply on
        {
            if (_driveState == state) //protect against another operation happening
            {
                int currentreading;
                if (state == DriveStates.DriveDistance)
                    currentreading = msg.Body.CurrentForeAftReading;
                else if (state == DriveStates.RotateDegrees)
                    currentreading = msg.Body.CurrentYawReading;
                else
                    throw new Exception();

                //if (power >= 0)
                //{
                    if (Math.Sign(power) * currentreading >= Math.Sign(power) * stopAt) //past mark, stop
                    {
                        //stop motors
                        drive.AllStop stop = new drive.AllStop();
                        stop.Body = new drive.AllStopRequest();
                        _mainPort.Post(stop);

                        responseport.Post(new DefaultUpdateResponseType());
                    }
                    else //keep driving
                    {
                        //re-enable listener
                        Activate(Arbiter.Receive<foreaftencoder.Replace>(false, encoderNotificationPort,
                            delegate(foreaftencoder.Replace replace)
                            {
                                MotorStopper(state, encoderNotificationPort, replace, stopAt, power, responseport);
                            }
                        ));
                    }
                //}
                //else
                //{
                //    if (currentreading <= stopAt) //driving backward, past mark
                //    {
                //        //stop motors
                //        drive.AllStop stop = new drive.AllStop();
                //        stop.Body = new drive.AllStopRequest();
                //        _mainPort.Post(stop);

                //        //SendNotification(_subMgrPort, new drive.Update(_state));
                //    }
                //    else //driving backward, not past mark
                //    {
                //        //re-enable listener
                //        Activate(Arbiter.Receive<foreaftencoder.Replace>(false, encoderNotificationPort,
                //            delegate(foreaftencoder.Replace replace)
                //            {
                //                MotorStopper(state, encoderNotificationPort, replace, stopAt, power);
                //            }
                //        ));
                //    }
                //}
            }
            else //_driveState != state
            {
                //post an error message on return port
                responseport.Post(new Fault());
            }
        }
           


        /// <summary>
        /// AllStop Handler
        /// </summary>
        /// <param name="update"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> AllStopHandler(drive.AllStop update)
        {
            _driveState = DriveStates.Stopped;
            _state.LeftWheel.MotorState.CurrentPower = 0;
            _state.RightWheel.MotorState.CurrentPower = 0;
            _state.TimeStamp = DateTime.Now;

            segway.DriveMessage drv = new segway.DriveMessage();
            drv.Velocity = 0;
            drv.Turn = 0;
            _segwayPort.Drive(drv);

            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

        #region HTTP Get / Post
        /// <summary>
        /// Http Get Handler.  Needed for XSLT transform
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> HttpGetHandler(HttpGet httpGet)
        {
            HttpListenerRequest request = httpGet.Body.Context.Request;
            HttpListenerResponse response = httpGet.Body.Context.Response;

            string path = request.Url.AbsolutePath;

            HttpResponseType rsp = new HttpResponseType(HttpStatusCode.OK, _state, _transform);
            httpGet.ResponsePort.Post(rsp);
            yield break;

        }

        /// <summary>
        /// Http Post Handler.  Handles http form inputs
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> HttpPostHandler(HttpPost httpPost)
        {
            // Use helper to read form data
            ReadFormData readForm = new ReadFormData(httpPost.Body.Context);
            _httpUtilities.Post(readForm);

            // Wait for result
            Activate(Arbiter.Choice(readForm.ResultPort,
                delegate(NameValueCollection parameters)
                {
                    if (parameters["StartDashboard"] != null)
                    {
                        string Dashboardcontract = "http://schemas.microsoft.com/robotics/2006/01/simpledashboard.html";
                        ServiceInfoType info = new ServiceInfoType(Dashboardcontract);
                        cons.Create create = new cons.Create(info);
                        create.TimeSpan = DsspOperation.DefaultShortTimeSpan;

                        ConstructorPort.Post(create);
                        Arbiter.Choice(
                            create.ResponsePort,
                            delegate(CreateResponse createResponse) { },
                            delegate(Fault f) { LogError(f); }
                        );
                    }
                    else if (parameters["EnableDrive"] != null)
                    {
                        drive.EnableDriveRequest req = new drive.EnableDriveRequest(true);
                        _mainPort.Post(new drive.EnableDrive(req));
                    }
                    else if (parameters["DisableDrive"] != null)
                    {
                        drive.EnableDriveRequest req = new drive.EnableDriveRequest(false);
                        _mainPort.Post(new drive.EnableDrive(req));
                    }
                    else if (parameters["AllStop"] != null)
                    {
                        _mainPort.Post(new drive.AllStop());
                    }
                    else if (parameters["DrivePower"] != null)
                    {
                        double power = 0;
                        double.TryParse(parameters["Power"], out power);
                        drive.SetDrivePowerRequest drivepower = new drive.SetDrivePowerRequest();
                        drivepower.LeftWheelPower = power;
                        drivepower.RightWheelPower = power;
                        _mainPort.Post(new drive.SetDrivePower(drivepower));
                    }
                    else if (parameters["DriveDistance"] != null)
                    {
                        double power = 0;
                        double.TryParse(parameters["Power"], out power);
                        double distance = 0;
                        double.TryParse(parameters["Distance"], out distance);
                        drive.DriveDistanceRequest drivedist = new drive.DriveDistanceRequest(distance, power);
                        _mainPort.Post(new drive.DriveDistance(drivedist));
                    }
                    else if (parameters["RotateAngle"] != null)
                    {
                        double power = 0;
                        double.TryParse(parameters["Power"], out power);
                        double angle = 0;
                        double.TryParse(parameters["Angle"], out angle);
                        drive.RotateDegreesRequest turn = new drive.RotateDegreesRequest(angle, power);
                        _mainPort.Post(new drive.RotateDegrees(turn));
                    }
                    else if (parameters["CustomPower"] != null)
                    {
                        double leftpower = 0;
                        double.TryParse(parameters["LeftPower"], out leftpower);
                        double rightpower = 0;
                        double.TryParse(parameters["RightPower"], out rightpower);
                        drive.SetDrivePowerRequest drivepower = new drive.SetDrivePowerRequest();
                        drivepower.LeftWheelPower = leftpower;
                        drivepower.RightWheelPower = rightpower;
                        _mainPort.Post(new drive.SetDrivePower(drivepower));
                    }
                    else
                    {
                        HttpPostFailure(httpPost, "Unknown Http Post");
                    }
                },
                delegate(Exception Failure)
                {
                    LogError(Failure.Message);
                })
            );
            yield break;
        }

        /// <summary>
        /// Send Http Post Success Response
        /// </summary>
        private void HttpPostSuccess(HttpPost httpPost)
        {
            HttpResponseType rsp =
                new HttpResponseType(HttpStatusCode.OK, _state, _transform);
            httpPost.ResponsePort.Post(rsp);
        }

        /// <summary>
        /// Send Http Post Failure Response
        /// </summary>
        private static void HttpPostFailure(HttpPost httpPost, string failureReason)
        {
            HttpResponseType rsp =
                new HttpResponseType(HttpStatusCode.ExpectationFailed, failureReason);
            httpPost.ResponsePort.Post(rsp);
        }
        #endregion

        /// <summary>
        /// Validate that the motors are configured and the drive is enabled.
        /// <remarks>Throws an exception (converted to fault) 
        /// when the service is not properly configured.</remarks>
        /// </summary>
        /// <param name="requireEncoders">validate encoder configuration</param>
        private void ValidateDriveConfiguration(bool requireEncoders)
        {
            if (_segwayPort == null)
            {
                LogError("The segway is not connected.");
                throw new InvalidOperationException();
            }

            if (!_state.IsEnabled)
            {
                LogError("The differential drive is not enabled.");
                throw new InvalidOperationException();
            }

            if (requireEncoders)
            {
                if (_foreaftEncoderPort == null || _leftrightEncoderPort == null)
                {
                    LogError("The wheel encoders are not properly configured");
                    throw new InvalidOperationException();
                }
            }
        }



    }
}
