//-----------------------------------------------------------------------
//  Segway left and right wheel Encoders
//
// Author: Ben Axelrod
// Date: April 2007
// Version 1.0
//
//-----------------------------------------------------------------------

using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.Core.DsspHttp;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;

using System;
using System.Net;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Security.Permissions;
using xml = System.Xml;
using W3C.Soap;

using submgr = Microsoft.Dss.Services.SubscriptionManager;
using segway = Robotics.SegwayBase.Proxy;


namespace Robotics.SegwayEncoder.LeftRight
{
    [Contract(Contract.Identifier)]
    [DisplayName("Segway Left and Right Wheel Encoders")]
    [Description("Provides access to the Segway RMP 200 wheel encoder.")]
    public class SegwayEncoderLeftRightService : DsspServiceBase
    {
        private SegwayEncoderLeftRightState _state = new SegwayEncoderLeftRightState();
        //private encoder.EncoderState _state = new encoder.EncoderState();

        [ServicePort("/SegwayEncoderLeftRight", AllowMultipleInstances = false)]
        private SegwayEncoderLeftRightOperations _mainPort = new SegwayEncoderLeftRightOperations();
        //private encoder.EncoderOperations _mainPort = new encoder.EncoderOperations();

        [Partner("Segway", Contract = segway.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate, Optional = false)]
        private segway.SegwayBaseOperations _segwayPort = new segway.SegwayBaseOperations();

        [Partner("SubMgr", Contract = submgr.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.CreateAlways, Optional = false)]
        private submgr.SubscriptionManagerPort _subMgrPort = new submgr.SubscriptionManagerPort();

        private bool _subscribed = false;

        Interleave _mainInterleave;

        public SegwayEncoderLeftRightService(DsspServiceCreationPort creationPort) : base(creationPort)
        { }

        protected override void Start()
        {
            //configure default state

            _state.CurrentLeftReading = 0;
            _state.CurrentRightReading = 0;
            _state.TicksPerMeter = 33215;

            #region base.Start()
            // Listen for each of my handlers
            _mainInterleave = ActivateDsspOperationHandlers();

            // Publish the service to the local Node Directory
            DirectoryInsert();

            // display HTTP service Uri
            LogInfo(LogGroups.Console, "Service uri: "); 
            #endregion

            // Subscribe to Segway for encoder notifications
            SubscribeToSegway();
        }

        /// <summary>
        /// Subscribe to appropriate data message update on segway
        /// </summary>
        private void SubscribeToSegway()
        {
            // Create a notification port
            segway.SegwayBaseOperations _notificationPort = new segway.SegwayBaseOperations();

            //create a custom subscription request
            segway.MySubscribeRequestType request = new segway.MySubscribeRequestType();

            //select appropriate data message
            request.DataMessageNumbers = new List<int>(1);
            request.DataMessageNumbers.Add(3); //WARNING: Magic number

            //Subscribe to the Segway and wait for a response
            Activate(
                Arbiter.Choice(_segwayPort.SelectiveSubscribe(request, _notificationPort),
                    delegate(SubscribeResponseType Rsp)
                    {
                        //update our state with subscription status
                        _subscribed = true;

                        LogInfo("SegwayEncoderLeftRight subscription success");

                        //Subscription was successful, start listening for sensor change notifications
                        _mainInterleave.CombineWith(new Interleave(
                            new ExclusiveReceiverGroup(
                                Arbiter.Receive<segway.Replace>(true, _notificationPort, SensorNotificationHandler)
                            ),
                            new ConcurrentReceiverGroup()
                        ));
                    },
                    delegate(Fault F)
                    {
                        LogError("SegwayEncoderLeftRight subscription failed");
                    }
                )
            );
        }


        /// <summary>
        /// Handle sensor update message from Segway
        /// </summary>
        public void SensorNotificationHandler(segway.Replace notify)
        {
            //update time
            _state.TimeStamp = DateTime.Now;

            _state.CurrentLeftReading = notify.Body.LeftWheelDisplacement;
            _state.CurrentRightReading = notify.Body.RightWheelDisplacement;

            //notify subscribers on any state change
            base.SendNotification(_subMgrPort, new Replace(_state));
        }


        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }




        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ResetHandler(Reset reset)
        {
            segway.ResetIntegratorBody msg = new segway.ResetIntegratorBody();
            //#define	RMP_PARAM_RIGHT_WHEEL_POS	1
            //#define	RMP_PARAM_LEFT_WHEEL_POS	2
            //#define	RMP_PARAM_YAW_POS			4
            //#define	RMP_PARAM_FORE_AFT_POS		8
            msg.Which = 3; //1 OR 2 = 3
            _segwayPort.ResetIntegrator(msg);
            reset.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }


        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReplaceHandler(Replace replace)
        {
            if (_subscribed)
                LogError("Already subscribed");
            else
            {
                _state = replace.Body;
                SaveState(_state);
                replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);
                SubscribeToSegway();
            }


            yield break;
        }

        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> SubscribeHandler(Subscribe subscribe)
        {
            yield return Arbiter.Choice(
                SubscribeHelper(_subMgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    SendNotification<Replace>(_subMgrPort, subscribe.Body.Subscriber, new Replace(_state));
                },
                delegate(Exception fault)
                {
                    LogError(fault);
                }
            );
        }

        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public IEnumerator<ITask> ReliableSubscribeHandler(ReliableSubscribe subscribe)
        {
            yield return Arbiter.Choice(
                SubscribeHelper(_subMgrPort, subscribe.Body, subscribe.ResponsePort),
                delegate(SuccessResult success)
                {
                    SendNotification<Replace>(_subMgrPort, subscribe.Body.Subscriber, new Replace(_state));
                },
                delegate(Exception fault)
                {
                    LogError(fault);
                }
            );
        }


    }



}

